"""
Model tools
"""
# Third party imports
from pyomo.core.base.var import Var
from pyomo.core.base.param import Param
from pyomo.core.base.set import BoundsInitializer
from pyomo.core.base.set import SetProduct
from pyomo.dae.contset import ContinuousSet
from pyomo.dae.diffvar import DerivativeVar

def get_vars(model):
    """Extract the variable information from the Pyomo model
    
    """
    vars_list = []
    model_Var = list(model.component_map(Var))
    model_dVar = list(model.component_map(DerivativeVar))
    vars_list = model_Var + model_dVar

    return vars_list

def get_vars_block(instance):
    """Alternative method for getting the model varialbes
    
    """
    model_variables = set()
    for block in instance.block_data_objects():
        block_map = block.component_map(Var)
        for name in block_map.keys():
            model_variables.add(name)
        
    return model_variables

def get_params(instance):
    """Get model params (to delete)
    
    """
    param_list = list(instance.component_map(Param))
    
    return param_list

def get_result_vars(model):
    """Get the Vars and Params needed for the results object
    
    """
    result_vars = get_vars(model)
    result_vars += get_params(model)
    
    return result_vars

def get_index_sets(model_var_obj):
    """Retuns a list of the index sets for the model variable
    
    Args:
        model (ConcreteModel): The pyomo model under investigation
        
        var (str): The name of the variable
        
    Returns:
        
        index_set (list): list of indecies
    
    """
    index_dict = {}
    index_set = []
    index = model_var_obj.index_set()
    if not isinstance(index, SetProduct):
        index_set.append(index)
    elif isinstance(index, SetProduct): # or isinstance(index, SetProduct):
        index_set.extend(index.subsets())
    else:
        return None
    
    return index_set

def index_set_info(index_list):
    """Returns whether index list contains a continuous set and where the
    index is
    
    Args:
        index_list (list): list of indicies produced by get_index_sets
        
    Returns:
        cont_set_info (tuple): (Bool, index of continuous set)
        
    """
    index_dict = {'cont_set': [],
                  'other_set': [],
                  }

    for i, index_set in enumerate(index_list):
        if isinstance(index_set, ContinuousSet):
            index_dict['cont_set'].append(i)
        else:
            index_dict['other_set'].append(i)
    index_dict['other_set'] = tuple(index_dict['other_set'])
        
    return index_dict

def change_continuous_set(cs, new_bounds):
    """Changes the bounds of the continuous set
    
    Args:
        cs (ContinuousSet): continuous set to change bound on
        
        new_bounds (tuple): New lower and upper bounds for cs
        
    Returns:
        None
    
    """
    cs.clear()
    cs._init_domain._set = None
    cs._init_domain._set = BoundsInitializer(new_bounds)
    domain = cs._init_domain(cs.parent_block(), None)
    cs._domain = domain
    domain.parent_component().construct()
    
    for bnd in cs.domain.bounds():
        if bnd is not None and bnd not in cs:
            cs.add(bnd)
    cs._fe = sorted(cs)
    
    return None


"""
Base Optimizer Class
"""
import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import (
    lsqr,
    spsolve,
    ) 
from scipy.optimize import least_squares

from pyomo.dae import *
from pyomo.environ import *
from pyomo.environ import (
    Suffix,
    )

from kipet.core_methods.PyomoSimulator import *
from kipet.core_methods.ResultsObject import *
from kipet.mixins.JumpsMixin import JumpsMixin


class Optimizer(JumpsMixin, PyomoSimulator):
    """Base optimizer class.

    Note:
        This class is not intended to be used directly by users

    Attributes:
        model (model): Pyomo model.

    """
    def __init__(self, model):
        """Optimizer constructor.

        Note: 
            Makes a shallow copy to the model. Changes applied to 
            the model within the simulator are applied to the original
            model passed to the simulator

        Args:
            model (Pyomo model)
        """
        super(Optimizer, self).__init__(model)
        
    def run_sim(self, solver, **kdws):
        raise NotImplementedError("Optimizer abstract method. Call child class")       

    def run_opt(self, solver, **kwds):
        raise NotImplementedError("Optimizer abstract method. Call child class")
        
    @staticmethod
    def add_warm_start_suffixes(model, use_k_aug=False):
        """Adds suffixed variables to problem"""
        
        # Ipopt bound multipliers (obtained from solution)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        # Ipopt bound multipliers (sent to solver)
        model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
        # Obtain dual solutions from first solve and send to warm start
        model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        
        if use_k_aug:
            model.dof_v = Suffix(direction=Suffix.EXPORT)
            model.rh_name = Suffix(direction=Suffix.IMPORT)
            
        return None
            
    @staticmethod
    def update_warm_start(model):
        """Updates the suffixed variables for a warmstart"""
        
        model.ipopt_zL_in.update(model.ipopt_zL_out)
        model.ipopt_zU_in.update(model.ipopt_zU_out)
        
        return None
    
    def _solve_S_from_DC(self, C_dataFrame, tee=False, with_bounds=False, max_iter=200):
        """Solves a basic least squares problems with SVD.
        
        Args:
            C_dataFrame (DataFrame) data frame with concentration values
        
        Returns:
            DataFrame with estimated S_values 
        """
        D_data = self.model.D
        if self._n_meas_lambdas:
            # build Dij vector
            D_vector = np.zeros(self._n_meas_times*self._n_meas_lambdas)
            
            row  = []
            col  = []
            data = []    
            for i,t in enumerate(self._meas_times):
                for j,l in enumerate(self._meas_lambdas):
                    for k,c in enumerate(self._mixture_components):
                        row.append(i*self._n_meas_lambdas+j)
                        col.append(j*self._n_components+k)
                        data.append(C_dataFrame[c][t])
                    D_vector[i*self._n_meas_lambdas+j] = D_data[t,l]    
                
                        
            Bd = coo_matrix((data, (row, col)),
                            shape=(self._n_meas_times*self._n_meas_lambdas,
                                   self._n_components*self._n_meas_lambdas)
                            )

            if not with_bounds:
                if self._n_meas_times == self._n_components:
                    s_array = spsolve(Bd, D_vector)
                elif self._n_meas_times>self._n_components:
                    result_ls = lsqr(Bd, D_vector,show=tee)
                    s_array = result_ls[0]
                else:
                    raise RuntimeError('Need n_t_meas >= self._n_components')
            else:
                nl = self._n_meas_lambdas
                nt = self._n_meas_times
                nc = self._n_components
                x0 = np.zeros(nl*nc)+1e-2
                M = Bd.tocsr()
                
                def F(x,M,rhs):
                    return  rhs-M.dot(x)

                def JF(x,M,rhs):
                    return -M

                if tee == True:
                    verbose = 2
                else:
                    verbose = 0
                    
                res_lsq = least_squares(F,x0,JF,
                                        bounds=(0.0,np.inf),
                                        max_nfev=max_iter,
                                        verbose=verbose,args=(M,D_vector))
                s_array = res_lsq.x
                
            s_shaped = s_array.reshape((self._n_meas_lambdas,self._n_components))
        else:
            s_shaped = np.empty((self._n_meas_lambdas,self._n_components))

        return s_shaped
    
    def run_lsq_given_P(self, solver, parameters, **kwds):
        """Gives a raw estimate of S given kinetic parameters.
        
        Args:
            solver (str): name of the nonlinear solver to used
          
            solver_opts (dict, optional): options passed to the nonlinear solver
        
            variances (dict, optional): map of component name to noise variance. The
            map also contains the device noise variance
            
            tee (bool,optional): flag to tell the optimizer whether to stream output
            to the terminal or not
            initialization (bool, optional): flag indicating whether result should be 
            loaded or not to the pyomo model
        
        Returns:
            Results object with loaded results
        """
        solver_opts = kwds.pop('solver_opts', dict())
        variances = kwds.pop('variances',dict())
        tee = kwds.pop('tee',False)
        initialization = kwds.pop('initialization',False)
        wb = kwds.pop('with_bounds',True)
        max_iter = kwds.pop('max_lsq_iter',200)
        
        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before runing simulation')

        #self.model =copy.deepcopy(self.model)

        base_values = ResultsObject()
        base_values.load_from_pyomo_model(self.model,
                                          to_load=['Z','dZdt','X','dXdt','Y'])
        
        # fixes parameters 
        old_values = {}        
        for k,v in parameters.items():
            if self.model.P[k].fixed ==False:
                old_values[k] = self.model.P[k].value
                self.model.P[k].value = v
                self.model.P[k].fixed = True

        for k,v in self.model.P.items():
            if not v.fixed:
                print('***WARNING parameter {} is not fixed. This method expects all parameters to be fixed.'.format(k))
            
        # deactivates objective functions for simulation                
        objectives_map = self.model.component_map(ctype=Objective,active=True)
        active_objectives_names = []
        for obj in objectives_map.values():
            name = obj.cname()
            active_objectives_names.append(name)
            obj.deactivate()

            
        opt = SolverFactory(solver)
        for key, val in solver_opts.items():
            opt.options[key]=val

        solver_results = opt.solve(self.model,tee=tee)

        #unfixes the parameters that were fixed
        for k,v in old_values.items():
            if not initialization:
                self.model.P[k].value = v 
            self.model.P[k].fixed = False
            self.model.P[k].stale = False
        # activates objective functions that were deactivated
        active_objectives_names = []
        objectives_map = self.model.component_map(ctype=Objective)
        for name in active_objectives_names:
            objectives_map[name].activate()

        # unstale variables that were marked stale
        for var in self.model.component_map(ctype=Var).values():
            if not isinstance(var, DerivativeVar):
                for var_data in var.values():
                    var_data.stale=False
            else:
                for var_data in var.values():
                    var_data.stale=True

        # retriving solutions to results object  
        results = ResultsObject()
        results.load_from_pyomo_model(self.model,
                                      to_load=['Z','dZdt','X','dXdt','Y'])

        c_array = np.zeros((self._n_allmeas_times,self._n_components))
        for i,t in enumerate(self._allmeas_times):
            for j,k in enumerate(self._mixture_components):
                c_array[i,j] = results.Z[k][t]

        results.C = pd.DataFrame(data=c_array,
                                 columns=self._mixture_components,
                                 index=self._allmeas_times)
        
        D_data = self.model.D
        
        if self._n_allmeas_times and self._n_allmeas_times<self._n_components:
            raise RuntimeError('Not enough measurements num_meas>= num_components')

        # solves over determined system
        s_array = self._solve_S_from_DC(results.C,
                                        tee=tee,
                                        with_bounds=wb,
                                        max_iter=max_iter)

        d_results = []
        for t in self._meas_times:
            for l in self._meas_lambdas:
                d_results.append(D_data[t,l])
        d_array = np.array(d_results).reshape((self._n_meas_times,self._n_meas_lambdas))
                        
        results.S = pd.DataFrame(data=s_array,
                                 columns=self._mixture_components,
                                 index=self._meas_lambdas)

        results.D = pd.DataFrame(data=d_array,
                                 columns=self._meas_lambdas,
                                 index=self._meas_times)        

        if initialization:
            for t in self.model.allmeas_times:
                for k in self.mixture_components:
                    self.model.C[t,k].value = self.model.Z[t,k].value

            for l in self.model.meas_lambdas:
                for k in self.mixture_components:
                    self.model.S[l,k].value =  results.S[k][l]
        else:
            if not base_values.Z.empty:
                self.initialize_from_trajectory('Z',base_values.Z)
                self.initialize_from_trajectory('dZdt',base_values.dZdt)
            if not base_values.X.empty:
                self.initialize_from_trajectory('X',base_values.X)
                self.initialize_from_trajectory('dXdt',base_values.dXdt)
        
        return results


"""
This module contains the code related to jumps - removes redundancy
"""
from pyomo.core import *
from pyomo.dae import *
from pyomo.environ import *

from kipet.mixins.VisitorMixins import ReplacementVisitor
    
def set_up_jumps(self, kwargs):

    var_dic = kwargs.pop("jump_states", None)
    jump_times = kwargs.pop("jump_times", None)
    feed_times = kwargs.pop("feed_times", None)

    self.disc_jump_v_dict = var_dic
    self.jump_times_dict = jump_times  # now dictionary
    self.feed_times_set = feed_times
    if not isinstance(self.disc_jump_v_dict, dict):
        print("disc_jump_v_dict is of type {}".format(type(self.disc_jump_v_dict)))
        raise Exception  # wrong type
    if not isinstance(self.jump_times_dict, dict):
        print("disc_jump_times is of type {}".format(type(self.jump_times_dict)))
        raise Exception  # wrong type
    count = 0
    for i in self.jump_times_dict.keys():
        for j in self.jump_times_dict[i].items():
            count += 1
    if len(self.feed_times_set) > count:
        raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                    "There are more time points in feed_times than jump_times provided.")
    self.load_discrete_jump()

def load_discrete_jump(self):
    self.jump = True

    zeit = None
    for i in self.model.component_objects(ContinuousSet):
        zeit = i
        break
    if zeit is None:
        raise Exception('No continuous_set')

    self.time_set = zeit.name

    tgt_cts = getattr(self.model, self.time_set)  ## please correct me (not necessary!)
    self.ncp = tgt_cts.get_discretization_info()['ncp']
    fe_l = tgt_cts.get_finite_elements()
    fe_list = [fe_l[i + 1] - fe_l[i] for i in range(0, len(fe_l) - 1)]

    for i in range(0, len(fe_list)):  # test whether integer elements
        self.jump_constraints(i)
  
# I want to change this spaghetti
def jump_constraints(self, fe):
    # type: (int) -> None
    """ Take the current state of variables of the initializing model at fe and load it into the tgt_model
    Note that this will skip fixed variables as a safeguard.

    Args:
        fe (int): The current finite element to be patched (tgt_model).
    """
    ###########################
    if not isinstance(fe, int):
        raise Exception  # wrong type
    ttgt = getattr(self.model, self.time_set)
    ##############################
    # Inclusion of discrete jumps: (CS)
    if self.jump:
        vs = ReplacementVisitor()  #: trick to replace variables
        kn = 0
        for ki in self.jump_times_dict.keys():
            if not isinstance(ki, str):
                print("ki is not str")
            vtjumpkeydict = self.jump_times_dict[ki]
            for l in vtjumpkeydict.keys():
                self.jump_time = vtjumpkeydict[l]
                # print('jumptime:',self.jump_time)
                self.jump_fe, self.jump_cp = self.fe_cp(ttgt, self.jump_time)
                if self.jump_time not in self.feed_times_set:
                    raise Exception("Error: Check feed time points in set feed_times and in jump_times again.\n"
                                    "They do not match.\n"
                                    "Jump_time is not included in feed_times.")
                # print('jump_el, el:',self.jump_fe, fe)
                if fe == self.jump_fe + 1:
                    #################################
                    for v in self.disc_jump_v_dict.keys():
                        if not isinstance(v, str):
                            print("v is not str")
                        vkeydict = self.disc_jump_v_dict[v]
                        for k in vkeydict.keys():
                            if k == l:  # Match in between two components of dictionaries
                                var = getattr(self.model, v)
                                dvar = getattr(self.model, "d" + v + "dt")
                                con_name = 'd' + v + 'dt_disc_eq'
                                con = getattr(self.model, con_name)

                                self.model.add_component(v + "_dummy_eq_" + str(kn), ConstraintList())
                                conlist = getattr(self.model, v + "_dummy_eq_" + str(kn))
                                varname = v + "_dummy_" + str(kn)
                                self.model.add_component(varname, Var([0]))  #: this is now indexed [0]
                                vdummy = getattr(self.model, varname)
                                vs.change_replacement(vdummy[0])   #: who is replacing.
                                # self.model.add_component(varname, Var())
                                # vdummy = getattr(self.model, varname)
                                jump_delta = vkeydict[k]
                                self.model.add_component(v + '_jumpdelta' + str(kn),
                                                         Param(initialize=jump_delta))
                                jump_param = getattr(self.model, v + '_jumpdelta' + str(kn))
                                if not isinstance(k, tuple):
                                    k = (k,)
                                exprjump = vdummy[0] - var[(self.jump_time,) + k] == jump_param  #: this cha
                                # exprjump = vdummy - var[(self.jump_time,) + k] == jump_param
                                self.model.add_component("jumpdelta_expr" + str(kn), Constraint(expr=exprjump))
                                for kcp in range(1, self.ncp + 1):
                                    curr_time = self.t_ij(ttgt, self.jump_fe + 1, kcp)
                                    if not isinstance(k, tuple):
                                        knew = (k,)
                                    else:
                                        knew = k
                                    idx = (curr_time,) + knew
                                    con[idx].deactivate()
                                    e = con[idx].expr
                                    suspect_var = e.args[0].args[1].args[0].args[0].args[1]  #: seems that
                                    # e = con[idx].expr.clone()
                                    # e.args[0].args[1] = vdummy
                                    # con[idx].set_value(e)
                                    vs.change_suspect(id(suspect_var))  #: who to replace
                                    e_new = vs.dfs_postorder_stack(e)  #: replace
                                    con[idx].set_value(e_new)
                                    conlist.add(con[idx].expr)
                kn = kn + 1

    @staticmethod
    def t_ij(self, time_set, i, j):
        # type: (ContinuousSet, int, int) -> float
        """Return the corresponding time(continuous set) based on the i-th finite element and j-th collocation point
        From the NMPC_MHE framework by @dthierry.
    
        Args:
            time_set (ContinuousSet): Parent Continuous set
            i (int): finite element
            j (int): collocation point
    
        Returns:
            float: Corresponding index of the ContinuousSet
        """
        if i < time_set.get_discretization_info()['nfe']:
            h = time_set.get_finite_elements()[i + 1] - time_set.get_finite_elements()[i]  #: This would work even for 1 fe
        else:
            h = time_set.get_finite_elements()[i] - time_set.get_finite_elements()[i - 1]  #: This would work even for 1 fe
        tau = time_set.get_discretization_info()['tau_points']
        fe = time_set.get_finite_elements()[i]
        time = fe + tau[j] * h
        return round(time, 6)
    
    @staticmethod
    def fe_cp(self, time_set, feedtime):
        """Return the corresponding fe and cp for a given time
         Args:
            time_set:
            t:
        
        type: (ContinuousSet, float) -> tuple
                 
        """
        fe_l = time_set.get_lower_element_boundary(feedtime)
        fe = None
        j = 0
        for i in time_set.get_finite_elements():
            if fe_l == i:
                fe = j
                break
            j += 1
        h = time_set.get_finite_elements()[1] - time_set.get_finite_elements()[0]
        tauh = [i * h for i in time_set.get_discretization_info()['tau_points']]
        j = 0  #: Watch out for LEGENDRE
        cp = None
        for i in tauh:
            if round(i + fe_l, 6) == feedtime:
                cp = j
                break
            j += 1
        return fe, cp


"""
Expression Classes
"""
# Standad library imports

# Third party imports
from pyomo.core.expr.numeric_expr import DivisionExpression, ProductExpression, NegationExpression
from pyomo.environ import (
    ConcreteModel,
    Objective,
    Var, 
    )
from pyomo.environ import units as pyo_units

# KIPET library imports
from kipet.common.VisitorClasses import ReplacementVisitor
from kipet.post_model_build.replacement import _update_expression

from kipet.dev_tools.display import Print

DEBUG = False

_print = Print(verbose=DEBUG)


class ExpressionBlock():
    
    """Class for general expression block classes"""
    
    def __init__(self,
                 exprs=None,
                 ):
        
        self.exprs = exprs
        self._title = 'EXPR'

    def __len__(self):
        
        return len(self.exprs)
        
    def display(self):
        
        if self.exprs is not None:
            print(f'{self._title} expressions:')
            for key, expr in self.exprs.items():
                print(f'{key}: {expr.expression.to_string()}')


    def display_units(self):

        margin = 8
        if self.exprs is not None:
            print(f'{self._title} units:')
            for key, expr in self.exprs.items():
                #print(f'  {key}: {expr.units}')
                print(f'{str(key).rjust(margin)} : {expr.units}')
    

class ODEExpressions(ExpressionBlock):
    
    """Class for ODE expressions"""
    
    def __init__(self,
                 ode_exprs=None,
                 ):
        
        super().__init__(ode_exprs)
        self._title = 'ODE'
        
        
class AEExpressions(ExpressionBlock):
    
    """Class for AE expressions"""
    
    def __init__(self,
                 alg_exprs=None,
                 ):
        
        super().__init__(alg_exprs)
        self._title = 'ALG'

class Expression():
    
    """Class for individual expressions"""
    
    def __init__(self,
                 name,
                 expression):
        
        self.name = name
        self.expression = expression
        self.units = None
        self.expression_orig = None
        
    def __str__(self):
        
        return self.expression.to_string()
    
    @property
    def show_units(self):
        return self.units.to_string()
    
    def _change_to_unit(self, c_mod, c_mod_new):
        """Method to remove the fixed parameters from the ConcreteModel
        TODO: move to a new expression class
        """
        var_dict = c_mod_new.var_dict
        expr_new = self.expression
        if self.expression_orig is not None:
            expr_new = self.expression_orig    
        for model_var, var_obj in var_dict.items():
            old_var = c_mod[model_var][1]
            new_var = var_dict[model_var]
            expr_new = _update_expression(expr_new, old_var, new_var)
        return expr_new

    def check_units(self):#, c_mod, c_mod_new):
        """Check the expr units by exchanging the real model with unit model
        components
        
        Args:
            key (str): component represented in ODE
            
            expr (Expression): Expression object of ODE
            
            c_mod (Comp): original Comp object used to declare the expressions
            
            c_mod_new (Comp_Check): dummy model with unit components
            
        Returns:
            pint_units (): Returns expression with unit model components
        
        """
        self.units = pyo_units.get_units(self.expression)
        return None

    def check_division(self, eps=1e-12):
        """Add a small amount to the numerator and denominator in a
        DivisionExpression to improve the numerics
        
        """
        expr = self.expression
        
        if isinstance(expr, DivisionExpression):
        
            ex1, ex2 = expr.args
            ex1 += eps
            ex2 += eps
            expr_new = ex1/ex2
            
            self.expression = expr_new
            self.expression_orig = expr
    
        return None
    
    #Consolidate the units per expression - avoid checking the individual units
    
    def check_term(self, term, convert_to):
        
        unit_term = pyo_units.get_units(term)
        _print(f'     Units: {unit_term} ==> {convert_to}\n')
        _changed_flag = False
        term_new = term
        
        if unit_term is not None:
            _print('Starting conversion and making the new term\n')
            term_new = pyo_units.convert(term, to_units=getattr(pyo_units, convert_to))
            
        return term_new
    
    def check_expression_units(self, convert_to=None, scalar=1):
        
        if convert_to is None:
            raise ValueError('You need to supply a conversion')
        
        expr = self.expression
        expr_new = 0
        
        _print(f'The ODE expression being handled is:\n     {expr}\n')
        _print(f'The type of expression being handled is:\n     {type(expr)}\n')
    
        if isinstance(expr, NegationExpression):
            scalar *= -1
    
        if isinstance(self.expression, (DivisionExpression, ProductExpression)):
            _print(f'The number of terms in this expression is: 1\n')
            
            term = self.check_term(expr, convert_to)
            expr_new = scalar*term
        
        
        else:
            _print(f'The number of terms in this expression is: {len(expr.args)}\n')
            
            for i, term in enumerate(expr.args):
                _print(f'({i+1}) The expression term considered here is\n     {term}\n')
                term = self.check_term(term, convert_to)
                expr_new += scalar*term
        
        _print('The new expression is:\n')
        _print(f'     {expr_new}')
            
        self.expression = expr_new
        self.units = getattr(pyo_units, convert_to)
             
        return None
        
            
            
    
    
    
    
    
    
    
    
    


"""
Top Level helper classes for KIPET
"""

class AttrDict(dict):

    "This class lets you use nested dicts like accessing attributes"
    
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)
    
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setitem__(self, key, item):
        
        if isinstance(item, dict):
            return super(AttrDict, self).__setitem__(key, AttrDict(item))
        else:
            return dict.__setitem__(self, key, item)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


class DosingPoint():
    """Small class to handle the dosing points in a cleaner manner"""
    
    def __init__(self, component, time, step):
        
        self.component = component
        self.time = time
        self.step = step
    
    def __repr__(self):
        return f'{self.component}: {self.time}, {self.step}'
    
    def __str__(self):
        return self.__repr__()
    
    @property
    def as_list(self):
        return [self.component, self.time, self.step]



"""
General scaling method for Kipet models
"""
# Standard library imports
import copy

# Third party imports
from pyomo.environ import Param

# Kipet library imports
from kipet.core_methods.PyomoSimulator import PyomoSimulator
from kipet.common.VisitorClasses import ScalingVisitor, ReplacementVisitor
from kipet.top_level.variable_names import VariableNames

__var = VariableNames()

def scale_models(models_input, k_vals, name=None):
    """Takes a model or dict of models and iterates through them to update the
    odes and parameter values
    
    """
    models_dict = copy.deepcopy(models_input)
    
    if not isinstance(models_dict, dict):
        key = 'model-1' if name is None else name
        models_dict = {key: models_dict}
    
    d_init = {}
    
    for model in models_dict.values():
        d_init_model = _scale_model(model, k_vals)
        d_init.update(d_init_model)

    return d_init, models_dict

def _scale_model(model, k_vals):
    """Scales an individual model and updates the initial parameter dict
    
    """
    scaled_bounds = {}    

    if not hasattr(model, __var.model_parameter_scaled):
        add_scaling_parameters(model, k_vals)
        scale_parameters(model)
     
        for k, v in getattr(model, __var.model_parameter).items():
            lb, ub = v.bounds
            
            lb = lb/getattr(model, __var.model_parameter_scaled)[k].value
            ub = ub/getattr(model, __var.model_parameter_scaled)[k].value
            
            if ub < 1:
                print('Bounds issue, pushing upper bound higher')
                ub = 1.1
            if lb > 1:
                print('Bounds issue, pushing lower bound lower')
                lb = 0.9
                
            scaled_bounds[k] = lb, ub
                
            model_params = getattr(model, __var.model_parameter)
            model_params[k].setlb(lb)
            model_params[k].setub(ub)
            model_params[k].unfix()
            model_params[k].set_value(1)
            
    parameter_dict = {k: (1, scaled_bounds[k]) for k in k_vals.keys() if k in model_params}
            
    return parameter_dict

def add_scaling_parameters(model, k_vals=None):
    
    print("The model has not been scaled and will now be scaled using K parameters")
    
    if k_vals is None:
    
        setattr(model, __var.model_parameter_scaled, Param(model.parameter_names,                                                                                                                                                            
                    initialize={k: v for k, v in getattr(model, __var.model_parameter).items()},
                    mutable=True,
                    default=1))
    else:
        setattr(model, __var.model_parameter_scaled, Param(model.parameter_names,                                                                                                                                                            
                    initialize={k: v for k, v in k_vals.items() if k in getattr(model, __var.model_parameter)},
                    mutable=True,
                    default=1))
    
    return None
        
def scale_parameters(model, k_vals=None):
    """If scaling, this multiplies the constants in model.K to each
    parameter in model.P.
    
    I am not sure if this is necessary and will look into its importance.
    """
    if not hasattr(model, __var.model_parameter_scaled):
        add_scaling_parameters(model, k_vals=k_vals)
        
    scale = {}

    for var in __var.model_vars + __var.rate_vars:
        if hasattr(model, var):
            for i in getattr(model, var):
                if var == __var.model_parameter:
                    scale[id(getattr(model, var)[i])] = getattr(model, __var.model_parameter_scaled)[i]
                    continue
                scale[id(getattr(model, var)[i])] = 1

    for k, v in getattr(model, __var.ode_constraints).items():
        scaled_expr = _scale_expression(v.body, scale)
        getattr(model, __var.ode_constraints)[k] = scaled_expr == 0

def remove_scaling(model, bounds=None):
    
    """You need to reset the bounds on the parameter too"""
    """EP problem: parameter bounds are not respected"""
    
    if not hasattr(model, __var.model_parameter_scaled):
        raise AttributeError('The model is not scaled')
        
    for param in getattr(model, __var.model_parameter).keys():   
        change_value = getattr(model, __var.model_parameter_scaled)[param].value
        getattr(model, __var.model_parameter)[param].set_value(getattr(model, __var.model_parameter_scaled)[param].value)
        
        for k, v in model.odes.items():
            ep_updated_expr = update_expression(v.body, getattr(model, __var.model_parameter_scaled)[param], 1)
            model.odes[k] = ep_updated_expr == 0
        
        del getattr(model, __var.model_parameter_scaled)[param]
        
    if bounds is not None:
        for param, bound in bounds.items():
            if param in getattr(model, __var.model_parameter):
                getattr(model, __var.model_parameter)[param].setlb(bound[0])
                getattr(model, __var.model_parameter)[param].setub(bound[1])
        
    return None
        
def _scale_expression(expr, scale):
    
    visitor = ScalingVisitor(scale)
    return visitor.dfs_postorder_stack(expr)

def update_expression(expr, replacement_param, change_value):
    """Takes the noparam_infon-estiambale parameter and replaces it with its intitial
    value
    
    Args:
        expr (pyomo constraint expr): the target ode constraint
        
        replacement_param (str): the non-estimable parameter to replace
        
        change_value (float): initial value for the above parameter
        
    Returns:
        new_expr (pyomo constraint expr): updated constraints with the
            desired parameter replaced with a float
    
    """
    visitor = ReplacementVisitor()
    visitor.change_replacement(change_value)
    visitor.change_suspect(id(replacement_param))
    new_expr = visitor.dfs_postorder_stack(expr)
    
    return new_expr




"""
ModelComponent Classes
"""
# Third party imports
import pint

# Kipet library imports
from kipet.model_components.units_handler import convert_single_dimension 

class ModelElement():
    
    """Abstract Class to handle the modeling components in KIPET"""
    
    def __init__(self, 
                 name=None,
                 class_=None, 
                 value=None,
                 units=None,
                 unit_base=None,
                 description=None,
                 pyomo_var=None,
                 model_var=None,
                 units_orig=None,
                 conversion_factor=1,
                 ): 
        
        self.name = self._check_name(name)
        self.class_ = class_
        self.value = value
        self.description = description
    
        self.pyomo_var = pyomo_var
        self.model_var = model_var
        self.unit_base = unit_base
        self.ur = unit_base.ur
        self.units = 1*self.ur('') if units is None else 1*self.ur(units)
        self.units_orig = self.ur(units_orig)
        self.use_orig_units = False
        self.conversion_factor = conversion_factor
        #self._check_scaling()

    def _check_scaling(self):
            
        #print(f'\nConverting {self.name} ################################################')
        quantity = 1 * self.units
        quantity.ito_base_units()
        
        quantity = convert_single_dimension(self.ur, quantity, self.unit_base.TIME_BASE, power_fixed=False)
        quantity = convert_single_dimension(self.ur, quantity, self.unit_base.VOLUME_BASE, power_fixed=True)
        
        if not self.units.u == quantity.u:
            
            margin = 8
            
            print(f'Converting {self.name}:')
            print('From'.rjust(margin) + f' : {self.value:e} {self.units.u}')
            print('To'.rjust(margin) + f' : {self.value*quantity.m:e} {quantity.u}')
            
            self.conversion_factor = quantity.m
            self.units = 1*quantity.units
            
            if self.value is not None:
                self.value = quantity.m*self.value
                
            if hasattr(self, 'bounds') and self.bounds is not None:
                bounds = list(self.bounds)
                if bounds[0] is not None:
                    bounds[0] *= self.conversion_factor
                if bounds[1] is not None:
                    bounds[1] *= self.conversion_factor
                self.bounds = (bounds) 

    # def convert_unit(self, u_orig, u_goal, scalar=1, power=1, both_powers=False):
   
    #     c1 = 1*self.ur(u_orig)
    #     c2 = 1*self.ur(u_goal)
        
    #     power2 = 1
    #     if both_powers:
    #         power2 = power
            
    #     con = (c1**power).to((c2**power2).u)/c2**power2 * (c2**power2).u/(c1**power).u
        
    #     return scalar * con
    
    # def convert_single_dimension(self, unit_to_convert, u_goal, power_fixed=False):
        
    #     u_g = self.ur(u_goal)
    #     orig_dim = {k.replace('[', '').replace(']', ''): v for k, v in dict(u_g.dimensionality).items()}
    #     units = {k: v for k, v in unit_to_convert._units.items()}
    #     dim_to_find = list(orig_dim.keys())[0]
    #     power = orig_dim[dim_to_find]
        
    #     # print(f'unit_to_convert: {unit_to_convert}')
    #     # print(f'dim_to_find: {dim_to_find}')
    #     # print(f'power {power}')
        
    #     for dims in units:
            
    #         s = self.ur(dims)
    #         s = list({k.replace('[', '').replace(']', ''): v for k, v in dict(s.dimensionality).items()})[0]
    #         # print(f's: {s}')
            
    #         if s == dim_to_find:
    #             if power_fixed and abs(units[dims]) == power:                                       
    #                 power = units[dims]
    #                 u_orig = dims
    #                 # print(dims, power)
    #                 con = self.convert_unit(u_orig, u_goal, power=abs(power))
    #                 # print(con)
    #                 con = con ** (power/abs(power))
    #                 new_unit = unit_to_convert * con
    #                 return new_unit
    #             elif not power_fixed:
    #                 power = units[dims]
    #                 u_orig = dims
    #                 # print(dims, power)
    #                 con = self.convert_unit(u_orig, u_goal, power=abs(power), both_powers=True)
    #                 # print(con)
    #                 con = con ** (power/abs(power))
    #                 new_unit = unit_to_convert * con
    #                 return new_unit
        
    #     return unit_to_convert
        
    def _check_name(self, name):
        """Check for valid attr names in the given string
        
        Args:
            name (str): given name for a python attribute
        
        Returns:
            checked_name (str): valid attribute name
            
        """
        string_replace_dict = {
            ' ': '_',
            '-': 'n',
            '+': 'p',
            '.': '_',
            }
        
        name = str(name)
        
        if name[0].isnumeric():
            name = 'y' + name
        
        for k, v in string_replace_dict.items():
            name = name.replace(k, v)
        
        return name
    
    @property
    def in_orig_units(self):
        """Convert the value of the element back into its original unit value
        
        """
        current = self.value*self.units
        original = current.to(self.units_orig.units)
        return original.m
    
class ModelAlgebraic(ModelElement):
    
    class_ = 'model_algebraic'
    
    def __init__(self,
                 name=None,
                 value=None,
                 bounds=(None, None),
                 units=None,
                 unit_base=None,
                 description=None,
                 data=None,
                 step=None,
                 pyomo_var=None,
                 model_var=None,
                 units_orig=None,
                 conversion_factor=1,
                 ):
    
        super().__init__(name, ModelComponent.class_, value, units,
                         unit_base, description, pyomo_var, model_var, units_orig, conversion_factor)
   
        self.bounds = bounds
        self.data = data
        self.step = step
   
    def __str__(self):
        
        
        margin = 25
        settings = f'ModelAlgebraic\n'
        
        for key in self.__dict__: #['name', 'class_', 'value', 'units']:
            if key == 'class_':
                continue
            settings += f'{str(key).rjust(margin)} : {getattr(self, key)}\n'
            
        return settings
        
    def __repr__(self):
        
        return f'ModelAlgebraic({self.name})'
    
    @property
    def lb(self):
        """Lower bound property"""
        return self.bounds[0]

    @property
    def ub(self):
        """Upper bound property"""
        return self.bounds[1]

    
class ModelComponent(ModelElement):
    """A simple class for holding component information"""
    
    class_ = 'model_component'
    
    def __init__(self,
                 name=None,
                 state=None,
                 value=None,
                 variance=None,
                 units=None,
                 unit_base=None,
                 known=True,
                 bounds=(None,None),
                 description=None,
                 absorbing=True,
                 pyomo_var=None,
                 model_var=None,   
                 units_orig=None,
                 conversion_factor=1,
                 ):
    
        super().__init__(name, ModelComponent.class_, value, units,
                         unit_base, description, pyomo_var, model_var, units_orig, conversion_factor)
        # component attributes
        self.variance = variance
        self.state = 'concentration'
        self.known = known
        self.bounds = bounds
        self.absorbing = absorbing
        
        #self._check_units()
        
    def __str__(self):
        
        margin = 25
        settings = f'ModelComponent\n'
        
        for key in self.__dict__: #['name', 'class_', 'value', 'units']:
            if key == 'class_':
                continue
            settings += f'{str(key).rjust(margin)} : {getattr(self, key)}\n'
            
        return settings
        
    def _check_units(self):
      
        if self.state == 'concentration':
            check_quantity = 1 * self.units
            if not check_quantity.check('[concentration]'):
                raise AttributeError(f'Concentration units incorrect for species {self.name}')
    
    
    def __repr__(self):
        
        return f'ModelComponent({self.name})'
    
    @property
    def lb(self):
        """Lower bound property"""
        return self.bounds[0]

    @property
    def ub(self):
        """Upper bound property"""
        return self.bounds[1]

    
    
class ModelState(ModelElement):
    """A simple class for holding non-component state information"""
    
    class_ = 'model_state'
    
    def __init__(self,
                 name=None,
                 state='state',
                 value=None,
                 variance=None,
                 units=None,
                 unit_base=None,
                 known=True,
                 bounds=(None, None),
                 description=None,
                 pyomo_var=None,
                 model_var=None,
                 units_orig=None,
                 conversion_factor=1,
                 ):
    
        super().__init__(name, ModelComponent.class_, value, units,
                         unit_base, description, pyomo_var, model_var, units_orig, conversion_factor)
        # component attributes
        self.variance = variance
        self.state = state
        self.known = known
        self.bounds = bounds
        
        #self._check_units()
        
    def __str__(self):
        
        margin = 25
        settings = f'ModelState\n'
        
        for key in self.__dict__: #['name', 'class_', 'value', 'units']:
            if key == 'class_':
                continue
            settings += f'{str(key).rjust(margin)} : {getattr(self, key)}\n'
            
        return settings
        
    def _check_units(self):
      
        if self.state == 'state':
            check_quantity = 1 * self.units
            if not check_quantity.check('[concentration]'):
                raise AttributeError(f'Concentration units incorrect for species {self.name}')
    
    
    def __repr__(self):
        
        return f'ModelState({self.name})'

    @property
    def lb(self):
        """Lower bound property"""
        return self.bounds[0]

    @property
    def ub(self):
        """Upper bound property"""
        return self.bounds[1]        


class ModelParameter(ModelElement):
    """A simple class for holding kinetic parameter data
    
    TODO: change init to value
    """

    class_ = 'model_parameter'

    def __init__(self,
                 name,
                 value=None,
                 units=None,
                 unit_base=None,
                 bounds=(None, None),
                 fixed=False,
                 variance=None,
                 description=None,
                 pyomo_var=None,
                 model_var=None,
                 units_orig=None,
                 conversion_factor=1,
                 ):
    
        super().__init__(name, ModelComponent.class_, value, units,
                         unit_base, description, pyomo_var, model_var, units_orig, conversion_factor)
        self.bounds = bounds
        self.fixed = fixed
        self.variance = variance
        
        if self.value is None:
            self.value = sum(self.bounds)/2
    
    def __str__(self):
        
        margin = 25
        settings = 'ModelParameter\n'
        
        for key in self.__dict__: #['name', 'class_', 'value', 'units']:
            if key == 'class_':
                continue
            settings += f'{str(key).rjust(margin)} : {getattr(self, key)}\n'
            
        return settings
        
    def __repr__(self):
        
        return f'ModelParameter({self.name})'
    
    @property
    def lb(self):
        """Lower bound property"""
        return self.bounds[0]

    @property
    def ub(self):
        """Upper bound property"""
        return self.bounds[1]
    
    
class ModelConstant(ModelElement):
    """A simple class for holding kinetic parameter data
    
    TODO: change init to value
    """

    class_ = 'model_constant'

    def __init__(self,
                 name,
                 value=None,
                 units=None,
                 unit_base=None,
                 description=None,
                 pyomo_var=None,
                 model_var=None,
                 units_orig=None,
                 conversion_factor=1,
                 ):
    
        super().__init__(name, ModelComponent.class_, value, units,
                         unit_base, description, pyomo_var, model_var, units_orig, conversion_factor)
        self._class_ = type(self)
    
    def __str__(self):
        
        margin = 25
        settings = 'ModelConstant\n'
        
        for key in self.__dict__: #['name', 'class_', 'value', 'units']:
            if key == 'class_':
                continue
            settings += f'{str(key).rjust(margin)} : {getattr(self, key)}\n'
            
        return settings
        
    def __repr__(self):
        
        return f'ModelConstant({self.name})'



"""
Common functions between PE and MEE
"""
import numpy as np
from pyomo.environ import (
    Suffix,
    )

from kipet.common.read_hessian import *

class PEMixins(object):

    def _set_up_reduced_hessian(self, model_obj, time_set, component_set, var_name, index):
        """Method to declare the reduced hessian suffix variables"""
    
        for t in time_set:
            for c in component_set:
                v = getattr(model_obj, var_name)[t, c]
                self._idx_to_variable[index] = v
                self.model.red_hessian[v] = index
                index += 1
       
        return index
                        
    
    def _order_k_aug_hessian(self, unordered_hessian, var_loc):
        """
        not meant to be used directly by users. Takes in the inverse of the reduced hessian
        outputted by k_aug and uses the rh_name to find the locations of the variables and then
        re-orders the hessian to be in a format where the other functions are able to compute the
        confidence intervals in a way similar to that utilized by sIpopt.
        """
        vlocsize = len(var_loc)
        n_vars = len(self._idx_to_variable)
        hessian = np.zeros((n_vars, n_vars))
        
        for i, vi in enumerate(self._idx_to_variable.values()):
            for j, vj in enumerate(self._idx_to_variable.values()):
                if n_vars == 1:
                    h = unordered_hessian
                    hessian[i, j] = h
                else:
                    h = unordered_hessian[(var_loc[vi]), (var_loc[vj])]
                    hessian[i, j] = h
        print(hessian.size, "hessian size")
        return hessian
    
    def _compute_residuals(self, model_obj, exp_index=None):
        """4
        Computes the square of residuals between the optimal solution (Z) and the concentration data (C)
        Note that this returns a matrix of time points X components and it has not been divided by sigma^2

        This method is not intended to be used by users directly
        """
        nt = self._n_meas_times
        nc = self._n_actual
        residuals = dict()
        
        conc_data = ['C', 'Cm']
        
        for model_var in conc_data:     
            if hasattr(model_obj, model_var) and getattr(model_obj, model_var) is not None:     
                for index, value in getattr(model_obj, model_var).items():
                    res_index = list(index)
                    if exp_index:
                        res_index.insert(0, exp_index)                 
                    res_index = tuple(res_index)             
                    residuals[res_index] = (value.value - model_obj.Z[index].value)**2 
             
        return residuals
    
    @staticmethod
    def _get_nparams(model_obj, isSkipFixed=True):
        """Returns the number of unfixed parameters"""
        
        nparams = 0

        for v in model_obj.P.values():
            if v.is_fixed():
                print(str(v) + '\has been skipped for covariance calculations')
                continue
            nparams += 1
            
        if hasattr(model_obj, 'Pinit'):
            for v in model_obj.Pinit.values():
                if isSkipFixed:
                    if v.is_fixed():
                        print(str(v) + '\has been skipped for covariance calculations')
                        continue
                nparams += 1

        return nparams
    
    def _variances_p_calc(self, hessian, variances):
        """Computes the covariance for post calculation anaylsis
        
        """        
        print(f'Var: {variances}')
        
        
        
        nparams = self._n_params

        H = hessian[-nparams:, :]
 
        B = self._compute_B_matrix(variances)
        Vd = self._compute_Vd_matrix(variances)
        
        print(f'B: {B.shape}')
        print(f'H: {H.shape}')
        print(f'Vd: {Vd.shape}')
        
        R = B.T @ H.T
        A = Vd @ R
        L = H @ B
        V_theta = (A.T @ L.T).T
        variances_p = np.diag(V_theta)
        
        if hasattr(self, '_eigredhess2file') and self._eigredhess2file==True:
            save_eig_red_hess(V_theta)
        
        return variances_p, V_theta

    @staticmethod
    def add_warm_start_suffixes(model, use_k_aug=False):
        """Adds suffixed variables to problem"""
        
        # Ipopt bound multipliers (obtained from solution)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        # Ipopt bound multipliers (sent to solver)
        model.ipopt_zL_in = Suffix(direction=Suffix.EXPORT)
        model.ipopt_zU_in = Suffix(direction=Suffix.EXPORT)
        # Obtain dual solutions from first solve and send to warm start
        model.dual = Suffix(direction=Suffix.IMPORT_EXPORT)
        
        if use_k_aug:
            model.dof_v = Suffix(direction=Suffix.EXPORT)
            model.rh_name = Suffix(direction=Suffix.IMPORT)
            
        return None
            
    @staticmethod
    def update_warm_start(model):
        """Updates the suffixed variables for a warmstart"""
        
        model.ipopt_zL_in.update(model.ipopt_zL_out)
        model.ipopt_zU_in.update(model.ipopt_zU_out)
        
        return None



"""
Additional Input Options for Variance and Parameter Estimation Classes
Place into Optimizer?
"""

def add_inputs(est_object, kwds):
    
    est_object.fixedtraj = kwds.pop("fixedtraj", False)
    est_object.fixedy = kwds.pop("fixedy", False)
    est_object.inputs_sub = kwds.pop("inputs_sub", None)
    est_object.yfix = kwds.pop("yfix", None)
    est_object.yfixtraj = kwds.pop("yfixtraj", None)
    trajectories = kwds.pop("trajectories", None)
    
    if est_object.inputs_sub != None:
        for k in est_object.inputs_sub.keys():
            if not isinstance(est_object.inputs_sub[k], list):
                print("wrong type for inputs_sub {}".format(type(est_object.inputs_sub[k])))
                # raise Exception
            for i in est_object.inputs_sub[k]:
                # print(est_object.inputs_sub[k])
                # print(i)
                if est_object.fixedtraj == True or est_object.fixedy == True:
                    if est_object.fixedtraj == True:
                        for j in est_object.yfixtraj.keys():
                            for l in est_object.yfixtraj[j]:
                                if i == l:
                                    # print('herel:fixedy', l)
                                    if not isinstance(est_object.yfixtraj[j], list):
                                        print("wrong type for yfixtraj {}".format(type(est_object.yfixtraj[j])))
                                    reft = trajectories[(k, i)]
                                    est_object.fix_from_trajectory(k, i, reft)
                    if est_object.fixedy == True:
                        for j in est_object.yfix.keys():
                            for l in est_object.yfix[j]:
                                if i == l:
                                    # print('herel:fixedy',l)
                                    if not isinstance(est_object.yfix[j], list):
                                        print("wrong type for yfix {}".format(type(est_object.yfix[j])))
                                    for key in est_object.model.alltime.value:
                                        vark = getattr(est_object.model, k)
                                        vark[key, i].set_value(key)
                                        vark[key, i].fix()  # since these are inputs we need to fix this
                else:
                    print("A trajectory or fixed input is missing for {}\n".format((k, i)))
    """/end inputs section"""
    
    return None




"""
General KIPET Model Settings

This should remove all of the fixed variable names across the KIPET software
"""

class VariableNames(object):
    
    """Names of the KIPET model variables"""
    
    def __init__(self):
    
        # Variables for KIPET models    
    
        self.model_parameter = 'P'
        self.model_parameter_scaled = 'K'
        self.concentration_measured = 'Cm'
        self.concentration_spectra = 'C'
        self.concentration_spectra_abs = 'Cs'
        self.concentration_model = 'Z'
        self.concentration_model_rate = f'd{self.concentration_model}dt'
        self.state = 'U'
        self.state_model = 'X'
        self.state_model_rate = f'd{self.state_model}dt'
        self.spectra_species = 'S'
        self.spectra_data = 'D'
        self.user_defined = 'UD'
        self.huplc_data = 'Dhat'
        self.smooth_parameter = 'Ps'
        self.algebraic = 'Y'
        self.unwanted_contribution = 'g'
        self.ode_constraints = 'odes'
        
        self.dosing_variable = 'Dose'
        self.dosing_component = 'd_var'
        
        self.time_step_change = 'time_step_change'
        self.step_variable = 'step'
        # Debug options
        
        self.model_constant = 'Const'
        
        self.DEBUG = False
        
        
    @property
    def time_dependent_variables(self):
        
        """
        ['Z', 'dZdt', 'S', 'C', 'X', 'dXdt', 'U', 'Y']
        """
        
        model_vars = [self.concentration_model,
                      self.concentration_model_rate,
                      self.spectra_species,
                      self.concentration_spectra,
                      self.state_model,
                      self.state_model_rate,
                      self.state,
                      self.algebraic,
                      ]
        
        return model_vars
        
    @property
    def modeled_states(self):
        """
        ['Z', 'dZdt', 'X', 'dXdt']
        """
        model_vars = [self.concentration_model,
                      self.concentration_model_rate,
                      self.state_model,
                      self.state_model_rate,
                      ]
        
        return model_vars
    
    @property
    def model_vars(self):
        """
        ['Z', 'X', 'P', 'Y', 'step', 'Const']
        """
        model_vars = [self.concentration_model,
                      #self.concentration_model_rate,
                      self.state_model,
                      #self.state_model_rate,
                      self.model_parameter,
                      self.algebraic,
                      self.step_variable,
                      self.model_constant,
                      ]
        
        return model_vars
    
    @property
    def rate_vars(self):
        
        model_vars = [
            self.concentration_model_rate,
            self.state_model_rate,
            ]
        
        return model_vars
    
    @property
    def plot_vars(self):
        
        model_vars = [
            self.concentration_model_rate,
            self.state_model_rate,
            self.algebraic,
            self.step_variable,
            self.spectra_species,
            ]
        
        return model_vars


"""
Unit conversion helper functions

@author: kevin
"""
import attr
from enum import Enum

from kipet.dev_tools.display import Print


DEBUG = False

_print = Print(verbose=DEBUG)

class ConversionType(Enum):
    
    volume = 1
    time = 2
    generic = 3
    
    
@attr.s
class UnitConversion:

    unit = attr.ib()
    dim = attr.ib()
    power = attr.ib()
    dim_power = attr.ib()
    

def convert_unit(unit_registry, u_orig, u_goal, scalar=1, power=1, both_powers=False, reverse_power=False):
   
    c1 = 1*unit_registry(u_orig)
    c2 = 1*unit_registry(u_goal)
    
    power2 = 1
    if both_powers:
        power2 = power
        
    if reverse_power:
        power2, power = power, power2
        
    con = (c1**power).to((c2**power2).u)/c2**power2 * (c2**power2).u/(c1**power).u
    
    return scalar * con

def convert_single_dimension(unit_registry, unit_from, unit_to, power_fixed=False):
    
    
    conversion = ConversionType.generic
    
    if isinstance(unit_to, str):
        unit_to = unit_registry(unit_to)
        is_volume = unit_to.check('[volume]')
    
    if isinstance(unit_from, str):
        unit_from = unit_registry(unit_from)
        
    if is_volume:
        conversion = ConversionType.volume
    
    # print(conversion)
    # print(unit_from)
    # print(unit_to)
    
    u_to = unit_to
    u_from = unit_from
    
    _print('#' * 25)
    _print(f'\nCan {unit_from} be converted to {u_to}?')
    
    dim = {}
    dim_to = {k.replace('[', '').replace(']', ''): v for k, v in dict(u_to.dimensionality).items()}
    dim_from = {k.replace('[', '').replace(']', ''): v for k, v in dict(u_from.dimensionality).items()}
    
    
    units_to = {k: v for k, v in u_to._units.items()}
    units_from = {k: v for k, v in u_from._units.items()}
    
    uc_to = UnitConversion(unit=list(units_to.keys())[0],
                           dim=list(dim_to.keys())[0],
                           power=list(units_to.values())[0],
                           dim_power=list(dim_to.values())[0]
                           
                           )
    
    # print(uc_to)
   
    new_unit = unit_from
   
    for unit in units_from:
       
        _print(f'\n\tLooking at {unit}:\n')
        
        dims = dict(unit_registry(unit).dimensionality)
        dims = {k.replace('[', '').replace(']', ''): v for k, v in dims.items()}

        try:
            uc_from = UnitConversion(unit=unit, 
                                 dim=list(dims.keys())[0], 
                                 power=units_from[unit],
                                 dim_power=dim_from[list(dims.keys())[0]]
                                 )
        except:
            _print('\tNo match for this unit found in the reduced set, quitting')
            continue

        _print(f'\t{uc_from}')
        _print(f'\t{uc_to}\n')
        
        
        if (unit_registry(uc_from.unit)**uc_from.power).check("[power]"):
            #print(f'Handling power unit: {(ur(uc_from.unit)**uc_from.power).check("[power]")}')
            continue
        
        if uc_from.unit == uc_to.unit:
            _print('The units are already equal...quitting')
            break
            
        if uc_from.dim == uc_to.dim:
            
            if conversion != ConversionType.volume:
                is_volume = (unit_registry(uc_from.unit)**uc_from.power).check("[volume]")
                if is_volume:    
                    continue
    
                reverse_power = False
                if uc_from.power > uc_to.power:
                    reverse_power = True
                
                power = uc_from.power
                con = convert_unit(unit_registry, unit, uc_to.unit, power=abs(power), reverse_power=reverse_power)
                con = con ** (power/abs(power))
                new_unit = unit_from * con
                #print(new_unit)
                
            elif conversion == ConversionType.volume:
                _print(unit_registry(f'{unit}**{uc_from.power}'))
                if unit_registry(f'{unit}**{abs(uc_from.power)}').check('[volume]'):
                    _print('Converting from volume to volume')
                    
                    power = 3 #if uc_from.power > 0 else -3
                    if abs(uc_from.power) == abs(uc_to.power):
                        power = uc_from.power
                    
                    reverse_power = False
                    if abs(uc_from.power) < abs(uc_to.power):
                        reverse_power = True
                    # reverse_power = False
                    
                    both = False
                    if abs(uc_from.power) == abs(uc_to.power):
                        both = True
                    
                    _print(f'convert_unit(ur, {unit}, {uc_to.unit}, power={power}, both_powers={both}, reverse_power={reverse_power}')
                    _print(power)
                    con = convert_unit(unit_registry, unit, uc_to.unit, power=power, both_powers=both, reverse_power=reverse_power)
                    con = con ** (uc_from.power/abs(uc_from.power))
                    _print(unit_from)
                    _print(con)
                    
                    new_unit = unit_from * con
                    _print(new_unit)
            
                else:
                    continue
      
    
    _print('\nFinished checking\nResults:')
    _print(f'\t{new_unit}')
    
    return new_unit


"""
Replacement functions

This module can be used to replace parameters in the models with constant
values followed by deletion of the parameters from the model.
"""
# Standard library imports
import copy
import numpy as np
import pandas as pd

# Third party imports
from pyomo.environ import *

# KIPET library imports
from kipet.common.VisitorClasses import ReplacementVisitor 
#from kipet.top_level.variable_names import VariableNames

class ParameterReplacer():
    
    def __init__(self,
                 models=None,
                 fix_parameters=None,
                 parameter_name='P',
                 expression_names=['odes'],
                 user_fixed=True,
                 ):
      
        #self.__var = VariableNames()
        
        self.models = models if models is not None else []
        self.fix_parameters = fix_parameters
        self.parameter_name = parameter_name
        self.expression_names = expression_names
        self.user_fixed = user_fixed

    def remove_fixed_vars(self):
        """Method to remove the fixed parameters from the ConcreteModel
        """
        for model in self.models:
    
            # Change to use the direct models
            #model = _model.p_model       
            p_var = getattr(model, self.parameter_name)
    
            # Check which parameters are fixed and add them to the set:
            params_to_fix = set([k for k, v in p_var.items() if v.fixed])
            
            # If the user defined parameters are also to be fixed:
            #if self.user_fixed:
            #    params_to_fix.add(set(self.fix_parameters))
        
            # Replace the parameters with numeric values and remove Var from model
            for param in params_to_fix:   
                parameter_to_change = param
                if parameter_to_change in p_var.keys():
                    change_value = p_var[param].value
                
                    for expression_var in self.expression_names:
                        e_var = getattr(model, expression_var)
        
                        for k, v in e_var.items(): 
                            ep_updated_expr = _update_expression(v.body, p_var[param], change_value)
                            e_var[k] = ep_updated_expr == 0
                
                    if hasattr(model, 'parameter_names'):
                        model.parameter_names.remove(param)
                    del p_var[param]
        
        return None
    
def _update_expression(expr, replacement_param, change_value):
    """Takes the non-estiambale parameter and replaces it with its intitial
    value
    
    Args:
        expr (pyomo constraint expr): the target ode constraint
        
        replacement_param (str): the non-estimable parameter to replace
        
        change_value (float): initial value for the above parameter
        
    Returns:
        new_expr (pyomo constraint expr): updated constraints with the
            desired parameter replaced with a float
    
    """
    visitor = ReplacementVisitor()
    visitor.change_replacement(change_value)
    visitor.change_suspect(id(replacement_param))
    new_expr = visitor.dfs_postorder_stack(expr)
    
    return new_expr


"""
ReactionModel class

This is a big wrapper class for most of the KIPET methods
"""
# Standard library imports
import collections
import copy
import os
import pathlib
import sys

# Third party imports
import numpy as np
import pandas as pd
from pyomo.environ import ConcreteModel, Set, Var
from pyomo.core.base.var import IndexedVar
from pyomo.dae.diffvar import DerivativeVar

# Kipet library imports
import kipet.core_methods.data_tools as data_tools
from kipet.core_methods.EstimationPotential import (
    replace_non_estimable_parameters,
    rhps_method,
    )
from kipet.core_methods.EstimabilityAnalyzer import EstimabilityAnalyzer
from kipet.core_methods.FESimulator import FESimulator
from kipet.core_methods.ParameterEstimator import ParameterEstimator
from kipet.core_methods.PyomoSimulator import PyomoSimulator
from kipet.core_methods.TemplateBuilder import TemplateBuilder
from kipet.core_methods.VarianceEstimator import VarianceEstimator
# from kipet.common.component_expression import get_unit_model
from kipet.common.model_funs import step_fun
from kipet.post_model_build.pyomo_model_tools import get_vars
from kipet.dev_tools.display import Print
from kipet.post_model_build.scaling import scale_models
from kipet.post_model_build.replacement import ParameterReplacer
from kipet.mixins.TopLevelMixins import WavelengthSelectionMixins

from kipet.top_level.data_component import (
    DataBlock, 
    DataSet,
    )
from kipet.top_level.spectral_handler import SpectralData
from kipet.top_level.element_blocks import (
    AlgebraicBlock,
    ComponentBlock,
    ConstantBlock, 
    ParameterBlock, 
    StateBlock,
    )
from kipet.top_level.expression import (
    AEExpressions,
    Expression,
    ODEExpressions,
    )
from kipet.top_level.helper import DosingPoint
from kipet.top_level.settings import (
    Settings, 
    )
from kipet.top_level.variable_names import VariableNames
# from kipet.common.VisitorClasses import FindingVisitor
from pyomo.environ import units as pyo_units

from kipet.model_components.units_handler import convert_single_dimension

__var = VariableNames()
DEBUG=__var.DEBUG
_print = Print(verbose=DEBUG)
model_components = ['parameters', 'components', 'states', 'algebraics', 'constants', 'steps']

class ReactionModel(WavelengthSelectionMixins):
    
    """This should consolidate all of the Kipet classes into a single class to
    enable a simpler framework for using the software. 
    
    """
    __var = VariableNames()
    
    def __init__(self, *args, **kwargs):
        
        
        # self.Parameter = Parameter
        # self.Component = Component
        
        self.name = kwargs.get('name', 'Model-1')
        self.model = None
        self.builder = TemplateBuilder()
        self.components = ComponentBlock()   
        self.parameters = ParameterBlock()
        self.constants = ConstantBlock()
        self.algebraics = AlgebraicBlock()
        self.states = StateBlock()
        self.datasets = DataBlock()
        self.data = {}
        self.spectra = None
        
        self.results_dict = {}
        self.settings = Settings(category='model')
        self.algebraic_variables = []
        self.variances = {}
        self.odes = ODEExpressions()
        self.algs = AEExpressions()
        self.odes_dict = {}
        self.algs_dict = {}
        self.__flag_odes_built = False
        self.__flag_algs_built = False
        self.custom_objective = None
        self.optimized = False
        self.dosing_var = None
        self.dosing_points = None
        self._has_dosing_points = False
        self._has_step_or_dosing = False
        self._has_non_absorbing_species = False
        self._var_to_fix_from_trajectory = []
        self._var_to_initialize_from_trajectory = []
        self._default_time_unit = 'seconds'
        self.unit_base = kwargs.get('unit_base', None)
        
        self._G_data = {'G_contribution': None, 'Z_in': dict(), 'St': dict()}
        self.__var = VariableNames()

        # self.components_used = set()
        self.component_map = {k: getattr(self, k) for k in ['components',
                                                            'states',
                                                            'parameters',
                                                            'algebraics',
                                                            'constants']}
        
    def __repr__(self):
        
        # m = 20
        
        # kipet_str = f'ReactionModel Object {self.name}:\n\n'
        # kipet_str += f'{"ODEs".rjust(m)} : {hasattr(self, "odes") and getattr(self, "odes") is not None}\n'
        # kipet_str += f'{"Algebraics".rjust(m)} : {hasattr(self, "odes") and getattr(self, "odes") is not None}\n'
        # kipet_str += f'{"Model".rjust(m)} : {hasattr(self, "model") and getattr(self, "model") is not None}\n'
        # kipet_str += '\n'
        
        # kipet_str += f'{self.components}\n'
        # kipet_str += f'Algebraic Variables:\n{", ".join([str(av) for av in self.algebraic_variables])}\n\n'
        # kipet_str += f'{self.parameters}\n'
        # kipet_str += f'{self.datasets}\n'
        
        return f'KipetModel {self.name}'#kipet_str
    
    def __str__(self):
        
        m = 20
        
        kipet_str = f'ReactionModel Object {self.name}:\n\n'
        kipet_str += f'{"ODEs".rjust(m)} : {hasattr(self, "odes") and getattr(self, "odes") is not None}\n'
        kipet_str += f'{"Algebraics".rjust(m)} : {hasattr(self, "odes") and getattr(self, "odes") is not None}\n'
        kipet_str += f'{"Model".rjust(m)} : {hasattr(self, "model") and getattr(self, "model") is not None}\n'
        kipet_str += '\n'
        
        kipet_str += f'{self.components}\n'
        kipet_str += f'Algebraic Variables:\n{", ".join([str(av) for av in self.algebraic_variables])}\n\n'
        kipet_str += f'{self.parameters}\n'
        kipet_str += f'{self.datasets}\n'
    
        return kipet_str
    
    """Model Components"""
    
    def _make_set_up_model(self):
        """Make the dummy model for initial pyomo vars
        
        """
        self._set_up_model = ConcreteModel()
        self._set_up_model.indx = Set(initialize=[0])
        return None
    
    def Component(self, name, index, units):
        """Creates the initial pyomo variables for model building
        
        """
        if not hasattr(self, '_set_up_model'):
            self._make_set_up_model()
        
        setattr(self._set_up_model, f'{name}_indx', Set(initialize=[0]))
        sets = [getattr(self._set_up_model, f'{name}_indx')]*index

        if units is None:
            p_units = str(1)
            comp_units = 1*self.unit_base.ur('')

        else:
            con = 1
            margin = 6
            print(f'Checking units for {name.rjust(margin)}: \t{units}')
            comp_units = convert_single_dimension(self.unit_base.ur, units, self.unit_base.TIME_BASE)
            con *= comp_units.m
            comp_units = convert_single_dimension(self.unit_base.ur, str(comp_units.units), self.unit_base.VOLUME_BASE)
            con *= comp_units.m
            p_units = con*getattr(pyo_units, str(comp_units.units))
            comp_units = con*comp_units.units
            if not self.unit_base.ur(units) == comp_units:
                print(f'    Converted to........: \t{comp_units.units}')
            
        setattr(self._set_up_model, name, Var(*sets, initialize=1, units=p_units))
        var = getattr(self._set_up_model, name)
        return var[tuple([0 for i in range(index)])], comp_units
    
    def _add_model_component(self, name, index, model_var, *args, **kwargs):
        """Generic method for adding modeling components to the ReactionModel
        
        """
        units = kwargs.get('units', None)
        value = kwargs.get('value', 1)
        kwargs['units_orig'] = units
        
        par, con = self.Component(args[0], index, units=units)
        #print(f'#### in the add model: {con}')
        
        kwargs['value'] = value*con.m
        kwargs['units'] = str(con.units)
        kwargs['conversion_factor'] = con.m
        
        if 'bounds' in kwargs:
        # if hasattr(key_comp, 'bounds') and key_comp.bounds is not None:
            
            bounds = list(kwargs.get('bounds', [0, 0]))
            if bounds[0] is not None:
                bounds[0] *= con.m
            if bounds[1] is not None:
                bounds[1] *= con.m
            kwargs['bounds'] = (bounds) 
        
        kwargs['unit_base'] = self.ub
        kwargs['pyomo_var'] = par
        kwargs['model_var'] = model_var
        getattr(self, f'{name}s').add_element(*args, **kwargs)
        return par
    
    def parameter(self, *args, **kwargs):
        """Create a parameter with a localized pyomo var
        
        """
        return self._add_model_component('parameter',
                                        1, 
                                        self.__var.model_parameter, 
                                        *args, 
                                        **kwargs)
        
    def component(self, *args, **kwargs):
        """Create a component with a localized pyomo var
        
        """
        return self._add_model_component('component', 
                                        2, 
                                        self.__var.concentration_model, 
                                        *args, 
                                        **kwargs)
    
    def state(self, *args, **kwargs):
        """Create a state with a localized pyomo var
        
        """
        return self._add_model_component('state', 
                                        2, 
                                        self.__var.state_model, 
                                        *args, 
                                        **kwargs)
    
    def constant(self, *args, **kwargs):
        """Create a constant with a localized pyomo var
        
        """
        return self._add_model_component('constant', 
                                        1, 
                                        self.__var.model_constant, 
                                        *args, 
                                        **kwargs)
        
    def algebraic(self, *args, **kwargs):
        """Create a algebraic variable with a localized pyomo var
        
        """
        print('Warning: This is deprecated and will be removed in a future version')
        
        return self._add_model_component('algebraic', 
                                        2, 
                                        self.__var.algebraic, 
                                        *args, 
                                        **kwargs)
    
    def fixed_state(self, *args, **kwargs):
        """Create a algebraic variable for fixed states
        
        """
        return self._add_model_component('algebraic', 
                                        2, 
                                        self.__var.algebraic, 
                                        *args, 
                                        **kwargs)
        
    
    # def variable(self, *args, **kwargs):
    #     """Create a generic variable with a localized pyomo var
          # This will be a generic method that others can use...
    #     """
    #     if kwargs.get('index', None) is None:
    #         raise ValueError('Custom variables require an index list')
        
    #     return self._add_model_component('custom', 
    #                                     kwargs['index'], 
    #                                     'V', 
    #                                     *args, 
    #                                     **kwargs)
    
    def step(self, name, *args, **kwargs):
        """Create a step variable with a localized pyomo var
        
        """
        self._has_step_or_dosing = True
        if not hasattr(self, '_step_list'):
            self._step_list = {}
            
        var_name = 'step'
        if name not in self._step_list:
            self._step_list[name] = [kwargs]
        else:
            self._step_list[name].append(kwargs)
            
        par, _ = self.Component(name, 2, None)
        if not hasattr(self, f'{var_name}s'):
            setattr(self, f'{var_name}s', {})
        getattr(self, f'{var_name}s')[name] = [self.__var.step_variable, par]
        return par
    
    """Dosing profiles"""
    
    def add_dosing_point(self, component, time, step):
        """Add a dosing point or several (check template for how this is handled)
        
        """
        conversion_dict = {'state': self.__var.state_model, 
                           'concentration': self.__var.concentration_model,
                           }
        if component not in self.components.names:
            raise ValueError('Invalid component name')
        dosing_point = DosingPoint(component, time, step)
        model_var = conversion_dict[self.components[component].state]
        if self.dosing_points is None:
            self.dosing_points = {}
        if model_var not in self.dosing_points.keys():
            self.dosing_points[model_var] = [dosing_point]
        else:
            self.dosing_points[model_var].append(dosing_point)
        self._has_dosing_points = True
        self._has_step_or_dosing = True
        
    def add_dosing(self, n_steps=1):
        """At the moment, this is needed to set up the dosing variable
        """
        self._has_step_or_dosing = True
        self._number_of_steps = n_steps
    
        return None
    
    def call_fe_factory(self):
        """Somewhat of a wrapper for this simulator method, but better"""

        self.simulator.call_fe_factory({
            self.__var.dosing_variable: [self.__var.dosing_component]},
            self.dosing_points)

        return None
    
    """Model data"""
    
    def add_data(self, *args, **kwargs):
        
        name = kwargs.get('name', None)
        time_scale = kwargs.get('time_scale', self.unit_base.TIME_BASE)
        
        time_conversion = 1
        if time_scale != self.unit_base.TIME_BASE:
            time_conversion = self.ub.ur(time_scale).to(self.unit_base.TIME_BASE).m
        
        if len(args) > 0:
            name = args[0]
        if name is None:
            name = f'ds{len(self.data) + 1}'
            
        filename = kwargs.get('file', None)
        data = kwargs.pop('data', None)
        category = kwargs.pop('category', None)
        remove_negatives = kwargs.get('remove_negatives', False)
    
        if filename is not None:
            calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
            filename = pathlib.Path(calling_file_name).joinpath(filename)
            kwargs['file'] = filename
            dataframe = data_tools.read_file(filename)
        elif filename is None and data is not None:
            dataframe = data
        else:
            raise ValueError('User must provide filename or dataframe')
        
        dataframe.index = dataframe.index*time_conversion
        
        if category == 'spectral':
            D_data = SpectralData(name, remove_negatives=remove_negatives)    
            D_data.add_data(dataframe)
            self.spectra = D_data
            return None
                    
        else:
            self.data[name] = dataframe
            dataset = DataSet(name,
                              category=category,
                              data = dataframe,
                              file = filename,
                              )
            
            if remove_negatives:
                dataset.remove_negatives()
            self.datasets.add_dataset(dataset)
        
        self._check_data_matches(name)
        self.datasets._check_duplicates()
            
        return None
    
    # Check for duplicates ==> DataBlock
    def _check_data_matches(self, name, name_is_from_model=False):
        """Easy data mapping to ElementBlocks"""
        # Reassignment
        
        # for name, dataset in self.data.items():
        #     print(name)
        #     for col in dataset.columns:
        #         print(col)
        #         for block in blocks:
        #             comp_set = getattr(self, block).names
        #             for comp in comp_set:
        #                 if col in comp_set:
        #                     print(block)
        #                     getattr(self, block)[col].data = dataset[col]
            
        blocks = ['components', 'states', 'algebraics']
        
        if not name_is_from_model:
            # matched_data_vars = set()
            # all_data_vars = set()
            dataset = self.datasets[name]
            if dataset.category in ['spectral', 'trajectory']:
                return None
            else:
                for block in blocks:
                    for col in dataset.data.columns:
                        #all_data_vars.add(col)
                        if col in getattr(self, block).names:
                            setattr(getattr(self, block)[col], 'data_link', dataset.name)
                            #matched_data_vars.add(col)
        
            #unmatched_vars = all_data_vars.difference(matched_data_vars)

        else:
            for block in blocks:
                if name in getattr(self, block).names:
                    for dataset in self.datasets:
                        for col in dataset.data.columns:
                            if col == name:
                                setattr(getattr(self, block)[name], 'data_link', dataset.name)

        return None

    """Template building"""

    def set_times(self, start_time=None, end_time=None):
        """Add times to model for simulation (overrides data-based times)"""
        
        if start_time is None or end_time is None:
            raise ValueError('Time needs to be a number')
        
        self.settings.general.simulation_times = (start_time, end_time)
        return None
    
    def _unwanted_G_initialization(self, *args, **kwargs):
        """Prepare the ParameterEstimator model for unwanted G contributions
        
        """
        self.builder.add_qr_bounds_init(bounds=(0,None),init=1.1)
        self.builder.add_g_bounds_init(bounds=(0,None))
        
        return None
    
    """ Expressions """
    
    def add_expression(self, name, expr, *args, **kwargs):
        """Add expressions (prev. algebraics) to the reaction model
        
        """        
        args_ = list(args)
        args_.insert(0, name)
        args = tuple(args_)
        
        # Adds algebraics anyways, for comparison purposes
        self._add_model_component('algebraic', 
                                  2, 
                                  self.__var.algebraic, 
                                  *args, 
                                  **kwargs)
        
        expr_ = Expression(name, expr)
        expr_.check_division()
        self.algs_dict.update(**{name: expr_})
        return expr
        
    def expression(self, label, expr):
        """
        Parameters
        ----------
        label : TYPE
            DESCRIPTION.
        expr : TYPE
            DESCRIPTION.

        Returns
        -------
        None.

        """
        expr = Expression(label, expr)
        if not hasattr(self, 'expr_dict'):
            self.expr_dict = {}
        self.expr_dict.update(**{label: expr})
        return None
    
    def add_ode(self, ode_var, expr):
        """Method to add an ode expression to the ReactionModel
        
        Args:
            ode_var (str): state variable
            
            expr (Expression): expression for rate equation as Pyomo Expression
            
        Returns:
            None
        
        """
        expr = Expression(ode_var, expr)
        self.odes_dict.update(**{ode_var: expr})
        
        return None
    
    def add_odes(self, ode_fun):
        """Takes in a dict of ODE expressions and sends them to add_ode
        
        Args:
            ode_fun (dict): dict of ode expressions
            
        Returns:
            None
        """
        if isinstance(ode_fun, dict):
            for key, value in ode_fun.items():
                self.add_ode(key, value)
        
        return None

    def _build_odes(self):
        """Builds the ODEs by passing them to the ODEExpressions class
        
        """
        odes = ODEExpressions(self.odes_dict)
        setattr(self, 'ode_obj', odes)
        self.odes = odes.exprs
        self.__flag_odes_built = True
        
        return None
        
    def add_algebraic(self, alg_var, expr):
        """Method to add an algebraic expression to the ReactionModel
        
        Args:
            alg_var (str): state variable
            
            expr (Expression): expression for algebraic equation as Pyomo
                Expression
            
        Returns:
            None
        
        """
        expr = Expression(alg_var, expr)
        expr.check_division()
        self.algs_dict.update(**{alg_var: expr})
    
        return None
    
    def add_algebraics(self, alg_fun):
        """Takes in a dict of algebraic expressions and sends them to
        add_algebraic
        
        Args:
            algebraics (dict): dict of algebraic expressions
            
        Returns:
            None
        """
        if isinstance(alg_fun, dict):
            for key, value in alg_fun.items():
                self.add_algebraic(key, value)
        
        return None
    
    def _build_algs(self):
        """Builds the algebraics by passing them to the AEExpressions class
        
        """ 
        algs = AEExpressions(self.algs_dict)
        setattr(self, 'alg_obj', algs)
        self.algs = algs.exprs
        self.__flag_algs_built = True

        return None
    
      
    def add_objective_from_algebraic(self, algebraic_var):
        """Declare an algebraic variable that is to be used in the objective
        
        TODO: add multiple alg vars for obj
        
        Args:
            algebraic_var (str): Variable representing the formula to be added
                to the objective
                
        Returns:
            None
            
        """
        self._check_data_matches(algebraic_var, name_is_from_model=True)
        self.custom_objective = algebraic_var
        
        return None
    
    def populate_template(self, *args, **kwargs):
        """Method handling all of the preparation for the TB
        
        """
        with_data = kwargs.get('with_data', True)
        
        # for model_component, obj in self.component_map.items():
            
        #     if len(obj) > 0:
        #         self.builder.add_model_element(obj)
                
        if len(self.states) > 0:
            self.builder.add_model_element(self.states)
        
        if len(self.algebraics) > 0:
            self.builder.add_model_element(self.algebraics)
        
        if len(self.components) > 0:
            self.builder.add_model_element(self.components)
        else:
            raise ValueError('The model has no components')
            
        if len(self.parameters) > 0:
            self.builder.add_model_element(self.parameters)
        else:
            self.allow_optimization = False   
        
        if len(self.constants) > 0:
            self.builder.add_model_element(self.constants)
        
        if with_data:
            if len(self.datasets) > 0 or self.spectra is not None:
                self.builder.input_data(self.datasets, self.spectra)
                self.allow_optimization = True
            elif len(self.datasets) == 0 and self.spectra is None:
                self.allow_optimization = False
            else:
                pass
            
        # Add the ODEs
        if len(self.odes_dict) != 0:
            self._build_odes()
            self.builder.set_odes_rule(self.odes)
        elif 'odes_bypass' in kwargs and kwargs['odes_bypass']:
            pass
        else:
            raise ValueError('The model requires a set of ODEs')
            
        if len(self.algs_dict) != 0:
            self._build_algs()
            if isinstance(self.algs, dict):
                self.builder.set_algebraics_rule(self.algs, asdict=True)
            else:
                self.builder.set_algebraics_rule(self.algs)
            
        if hasattr(self, 'custom_objective') and self.custom_objective is not None:
            self.builder.set_objective_rule(self.custom_objective)
        
        self.builder.set_parameter_scaling(self.settings.general.scale_parameters)
        self.builder.add_state_variance(self.components.variances)
        
        # if self._has_step_or_dosing:
        #     self.builder.add_dosing_var(self._number_of_steps)
        
        if self._has_dosing_points:
            self._add_feed_times()
            
        if hasattr(self, '_step_list') and len(self._step_list) > 0:
            self.builder.add_step_vars(self._step_list)
            
        # It seems this is repetitive - refactor
        if hasattr(self, '_G_contribution') and self._G_contribution is not None:
        #if self.settings.parameter_estimator.G_contribution is not None:
            self._unwanted_G_initialization()
            self.builder._G_contribution = self._G_contribution
        else:
            self.builder._G_contribution = None
        
        start_time, end_time = None, None
        if self.settings.general.simulation_times is not None:
            #print(f'times are: {type(self.settings.general.simulation_times)}')
            start_time, end_time = self.settings.general.simulation_times
       
        return start_time, end_time
    
    def _make_c_dict(self):
        
        model_components = ['parameters', 'components', 'states', 'algebraics', 'constants']
        
        if not hasattr(self, 'c'):
            self.c = dict()
            for mc in model_components: 
                if hasattr(self, f'{mc}') and len(getattr(self, f'{mc}')) > 0:
                    for comp in getattr(self, f'{mc}'):
                        #print(comp)
                        #print(comp.model_var, comp.pyomo_var)
                        self.c.update({comp.name: [comp.model_var, comp.pyomo_var]})
    
            if hasattr(self, 'steps'):
                self.c.update(getattr(self, 'steps'))
    
    def create_pyomo_model(self, *args, **kwargs):
        """Adds the component, parameter, data, and odes to the TemplateBuilder
        instance and creates the model. The model is stored under self.model
        and there is nothing returned.

        Args:
            None

        Returns:
            None

        """
        if hasattr(self, 'model'):
            del self.model
            
        skip_non_abs = kwargs.pop('skip_non_abs', False)
            
        start_time, end_time = self.populate_template(*args, **kwargs)
        self._make_c_dict()
        setattr(self.builder, 'c_mod', self.c)
        
        self.model = self.builder.create_pyomo_model(start_time, end_time)
        
        non_abs_comp = self.components.get_match('absorbing', False)
        
        if not skip_non_abs and len(non_abs_comp) > 0:
            self.builder.set_non_absorbing_species(self.model, non_abs_comp, check=True)    
        
        if hasattr(self,'fixed_params') and len(self.fixed_params) > 0:
            for param in self.fixed_params:
                self.model.P[param].fix()
            
        return None
    
    def _add_feed_times(self):
        
        feed_times = set()
        
        for model_var, dp in self.dosing_points.items():
            for point in dp:
                feed_times.add(point.time)
        
        self.builder.add_feed_times(list(feed_times))
        return None
    
    def _from_trajectories(self, estimator):
        """This handles all of the fixing, initializing, and scaling from 
        trajectory data
        
        """
        self._var_to_fix_from_trajectory = self.algebraics.fixed
        
        if len(self._var_to_fix_from_trajectory) > 0:
            for fix in self._var_to_fix_from_trajectory:
                if isinstance(fix[2], str):
                    if fix[2] in self.datasets.names:
                        fix[2] = self.datasets[fix[2]].data
                getattr(self, estimator).fix_from_trajectory(*fix)
                
        if len(self._var_to_initialize_from_trajectory) > 0:
            for init in self._var_to_initialize_from_trajectory:
                if isinstance(init[1], str):
                    if init[1] in self.datasets.names:
                        init[1] = self.datasets[init[1]].data
                getattr(self, estimator).initialize_from_trajectory(*init)
                
        return None
    
    """Simulation"""
    
    def simulate(self):
        """This should try to handle all of the simulation cases"""
    
        # Create the simulator object
        _print('Making simulator')
        self.create_simulator()
        # Add any previous trajectories, if given
        _print('Adding traj')
        self._from_trajectories('simulator')
        # Run the simulation
        _print('Running sim')
        self.run_simulation()
        
        return None
    
    def create_simulator(self):
        """This should try to handle all of the simulation cases"""
        
        _print('Setting up simulator:')
        sim_set_up_options = copy.copy(self.settings.simulator)
        _print(sim_set_up_options)
        dis_method = sim_set_up_options.pop('method', 'dae.collocation')
        
        if self._has_step_or_dosing:
            dis_method = 'fe'
        
        _print(dis_method)
        
        if dis_method == 'fe':
            simulation_class = FESimulator
        else:
            simulation_class = PyomoSimulator
        
        if self.model is None:
            # with_data her
            self.create_pyomo_model(*self.settings.general.simulation_times)
        
        self.s_model = self.model.clone()
        
        if hasattr(self.s_model, self.__var.model_parameter):
            for param in getattr(self.s_model, self.__var.model_parameter).values():
                param.fix()
        
        _print(simulation_class)
        
        simulator = simulation_class(self.s_model)
        _print('After class made')
        simulator.apply_discretization(self.settings.collocation.method,
                                       ncp=self.settings.collocation.ncp,
                                       nfe=self.settings.collocation.nfe,
                                       scheme=self.settings.collocation.scheme)
        
        _print('After disc')
        
        if self._has_step_or_dosing:
            for time in simulator.model.alltime.data():
                getattr(simulator.model, self.__var.dosing_variable)[time, self.__var.dosing_component].set_value(time)
                getattr(simulator.model, self.__var.dosing_variable)[time, self.__var.dosing_component].fix()
        
            #getattr(simulator.model, 'time_step_change')[0].fix()
        
        self.simulator = simulator
        print('Finished creating simulator')
        
        return None
        
    def run_simulation(self):
        """Runs the simulations, may be combined with the above at a later date
        
        """
        if self._has_dosing_points or isinstance(self.simulator, FESimulator):
            self.call_fe_factory()
        
        simulator_options = self.settings.simulator
        simulator_options.pop('method', None)
        self.results = self.simulator.run_sim(**simulator_options)
#        self.results.file_dir = self.settings.general.charts_directory
    
        return None
    
    # def reduce_spectra_data_set(self, dropout=4):
    #     """To reduce the computational burden, this can be used to reduce 
    #     the amount of spectral data used
        
    #     """
    #     A_set = [l for i, l in enumerate(self.model.meas_lambdas) if (i % dropout == 0)]
    #     return A_set
    
    def bound_profile(self, var, bounds):
        """Wrapper for TemplateBuilder bound_profile method"""
        
        self.builder.bound_profile(var=var, bounds=bounds)
        return None
    
    """Estimators"""
    
    def create_variance_estimator(self, **kwargs):
        """This is a wrapper for creating the VarianceEstimator"""
        
        if len(kwargs) == 0:
            kwargs = self.settings.collocation
        
        if self.model is None:    
            self.create_pyomo_model()  
        
        self.create_estimator(estimator='v_estimator', **kwargs)
        self._from_trajectories('v_estimator')
        return None
        
    def create_parameter_estimator(self, **kwargs):
        """This is a wrapper for creating the ParameterEstiamtor"""
        
        if len(kwargs) == 0:
            kwargs = self.settings.collocation
            
        if self.model is None:    
            self.create_pyomo_model()  
            
        self.create_estimator(estimator='p_estimator', **kwargs)
        self._from_trajectories('p_estimator')
        return None
        
    def initialize_from_simulation(self, estimator='p_estimator'):
        
        if not hasattr(self, 's_model'):
            _print('Starting simulation for initialization')
            self.simulate()
            _print('Finished simulation, updating variables...')

        _print(f'The model has the following variables:\n{get_vars(self.s_model)}')
        vars_to_init = get_vars(self.s_model)
        
        _print(f'The vars_to_init: {vars_to_init}')
        for var in vars_to_init:
            if hasattr(self.results, var):    
                _print(f'Updating variable: {var}')
                getattr(self, estimator).initialize_from_trajectory(var, getattr(self.results, var))
            else:
                continue
        
        return None
    
    def create_estimator(self, estimator=None):
        """This function handles creating the Estimator object
        
        Args:
            estimator (str): p_estimator or v_estimator for the PE or VE
            
        Returns:
            None
            
        """        
        if estimator == 'v_estimator':
            Estimator = VarianceEstimator
            est_str = 'VarianceEstimator'
            
        elif estimator == 'p_estimator':
            Estimator = ParameterEstimator
            est_str = 'ParameterEstimator'
            
        else:
            raise ValueError('Keyword argument estimator must be p_estimator or v_estimator.')  
        
        model_to_clone = self.model
        
        setattr(self, f'{estimator[0]}_model', model_to_clone.clone())
        setattr(self, estimator, Estimator(getattr(self, f'{estimator[0]}_model')))
        getattr(self, estimator).apply_discretization(self.settings.collocation.method,
                                                      ncp=self.settings.collocation.ncp,
                                                      nfe=self.settings.collocation.nfe,
                                                      scheme=self.settings.collocation.scheme)
        _print('Starting from_traj')
        self._from_trajectories(estimator)
        
        _print('Starting sim init')
        if self.settings.parameter_estimator.sim_init and estimator == 'p_estimator':
            self.initialize_from_simulation(estimator=estimator)
        
        if self._has_step_or_dosing:
            for time in getattr(self, estimator).model.alltime.data():
                getattr(getattr(self, estimator).model, self.__var.dosing_variable)[time, self.__var.dosing_component].set_value(time)
                getattr(getattr(self, estimator).model, self.__var.dosing_variable)[time, self.__var.dosing_component].fix()
        
        
        return None
    
    # def solve_variance_given_delta(self):
    #     """Wrapper for this VarianceEstimator function"""
    #     variances = self.v_estimator.solve_sigma_given_delta(**self.settings.variance_estimator)
    #     return variances
        
    def run_ve_opt(self):
        """Wrapper for run_opt method in VarianceEstimator"""
        
        if self.settings.variance_estimator.method == 'direct_sigmas':
            worst_case_device_var = self.v_estimator.solve_max_device_variance(**self.settings.variance_estimator)
            self.settings.variance_estimator.device_range = (self.settings.variance_estimator.best_accuracy, worst_case_device_var)
            
        self._run_opt('v_estimator', **self.settings.variance_estimator)
        
        return None
    
    def run_pe_opt(self):
        """Wrapper for run_opt method in ParameterEstimator"""
        if hasattr(self, '_G_data') and self._G_data is not None:
            pe_settings = {**self.settings.parameter_estimator, **self._G_data}
        else:
            pe_settings = {**self.settings.parameter_estimator} #, **self._G_data}, 
            
        self._run_opt('p_estimator', **pe_settings)
        
        return None
    
    def _update_related_settings(self):
        """Checks if conflicting options are present and fixes them accordingly
        
        """
        # Start with what is known
        if self.settings.parameter_estimator['covariance']:
            if self.settings.parameter_estimator['solver'] not in ['k_aug', 'ipopt_sens']:
                raise ValueError('Solver must be k_aug or ipopt_sens for covariance matrix')
        
        # If using sensitivity
        # solvers switch covariance to True
        if self.settings.parameter_estimator['solver'] in ['k_aug', 'ipopt_sens']:
            self.settings.parameter_estimator['covariance'] = True
        
        #Subset of lambdas
        if self.settings.variance_estimator['freq_subset_lambdas'] is not None:
            if isinstance(self.settings.variance_estimator['freq_subset_lambdas'], int):
                self.settings.variance_estimator['subset_lambdas' ] = self.reduce_spectra_data_set(self.settings.variance_estimator['freq_subset_lambdas']) 
        
        if self.settings.general.scale_pe and not self.settings.general.no_user_scaling:
            self.settings.solver.nlp_scaling_method = 'user-scaling'
    
        if self.settings.variance_estimator.max_device_variance:
            self.settings.parameter_estimator.model_variance = False
    
        return None
    
    def fix_parameter(self, param_to_fix):
        """Fixes parameter passed in as a list
        
        Args:
            param_to_fix (list): List of parameter names to fix
        
        Returns:
            None
        
        """
        if not hasattr(self, 'fixed_params'):
            self.fixed_params = []
        
        if isinstance(param_to_fix, str):
            param_to_fix = [param_to_fix]
            
        self.fixed_params += [p for p in param_to_fix]
        
        return None
    
    def run_opt(self):
        """Run ParameterEstimator but checking for variances - this should
        remove the VarianceEstimator being required to be implemented by the user
        
        """
        _print('Starting the RunOpt method')
        
        # Make the model if not present
        if self.model is None:    
            self.create_pyomo_model()  
            _print('Generating model')
        
        # Check if all needed data for optimization available
        if not self.allow_optimization:
            raise ValueError('The model is incomplete for parameter optimization')
            
        # Some settings are required together, this method checks this
        self._update_related_settings()
        
        # Check if all component variances are given; if not run VarianceEstimator
        has_spectral_data = self.spectra is not None
        has_all_variances = self.components.has_all_variances
        variances_with_delta = None
        
        # Check for bad options
        if self.settings.variance_estimator.method == 'direct_sigmas':
            raise ValueError('This variance method is not intended for use in the manner: see Ex_13_direct_sigma_variances.py')
        
        # If not all component variances are provided and spectral data is
        # present, the VE needs to be run
        if not has_all_variances and has_spectral_data:
            """If the data is spectral and not all variances are provided, VE needs to be run"""
            
            # Create the VE
            self.create_estimator(estimator='v_estimator')
            self.settings.variance_estimator.solver_opts = self.settings.solver
            # Optional max device variance
            if self.settings.variance_estimator.max_device_variance:
                max_device_variance = self.v_estimator.solve_max_device_variance(**self.settings.variance_estimator)
            
            # elif self.settings.variance_estimator.use_delta:
            #     variances_with_delta = self.solve_variance_given_delta()

            else:
                self.run_ve_opt()
                
                #print('Finished VE Opt - moving to PE opt')
        # If not a spectral problem and not all variances are provided, they
        # set to 1
        elif not has_all_variances and not has_spectral_data:
            for comp in self.components:
                try:
                    comp.variance = self.variances[comp.name]
                except:
                    comp.variance = 1
                
        # Create ParameterEstimator
        _print('Making PEstimator')
        self.create_estimator(estimator='p_estimator')
        
        variances = self.components.variances
        self.variances = variances
        
        # The VE results can be used to initialize the PE
        if 'v_estimator' in self.results_dict:
            if self.settings.general.initialize_pe:
                # Update PE using VE results
                self.initialize_from_variance_trajectory()
                # No initialization from simulation is needed
                self.settings.parameter_estimator.sim_init = False
 
            if self.settings.general.scale_pe:
                # Scale variables from VE results
                self.scale_variables_from_variance_trajectory()
            
            # Extract vairances
            self.variances = self.results_dict['v_estimator'].sigma_sq
        
        # If using max device variance, the variances are calculated differently
        elif self.settings.variance_estimator.max_device_variance:
            self.variances = max_device_variance
        
        # Under construction
        """ TODO
        # elif variances_with_delta is not None: 
        #     variances = variances_with_delta
        """
        # Optional variance scaling
        if self.settings.general.scale_variances:
            self.variances = self._scale_variances(self.variances)
        
        # Update PE solver settings and variances
        self.settings.parameter_estimator.solver_opts = self.settings.solver
        self.settings.parameter_estimator.variances = self.variances
        
        # Run the PE
        self.run_pe_opt()
        
        # Save results in the results_dict
        self.results = self.results_dict['p_estimator']
        self.results.file_dir = pathlib.Path.cwd() #self.settings.general.charts_directory
        
        # Tells MEE that the individual model is already solved
        self.optimized = True
        
        return self.results
    
    @staticmethod
    def _scale_variances(variances):
        
        max_var = max(variances.values())
        scaled_vars = {comp: var/max_var for comp, var in variances.items()}
        return scaled_vars

    def _run_opt(self, estimator, *args, **kwargs):
        """Runs the respective optimization for the estimator"""
        
        if not hasattr(self, estimator):
            raise AttributeError(f'ReactionModel has no attribute {estimator}')
            
        self.results_dict[estimator] = getattr(self, estimator).run_opt(*args, **kwargs)
        return self.results_dict[estimator]
    
    def initialize_from_variance_trajectory(self, variable=None, obj='p_estimator'):
        """Wrapper for the initialize_from_trajectory method in
        ParameterEstimator
        
        """
        source = self.results_dict['v_estimator']
        self._from_trajectory('initialize', variable, source, obj)
        return None
    
    def initialize_from_trajectory(self, variable_name=None, source=None):
        """Wrapper for the initialize_from_trajectory method in
        ParameterEstimator or PyomoSimulator
        
        """
        self._var_to_initialize_from_trajectory.append([variable_name, source])
        return None
    
    def scale_variables_from_variance_trajectory(self, variable=None, obj='p_estimator'):
        """Wrapper for the scale_varialbes_from_trajectory method in
        ParameterEstimator
        
        """
        source = self.results_dict['v_estimator']
        self._from_trajectory('scale_variables', variable, source, obj)
        return None
        
    @staticmethod
    def _get_source_data(source, var):
        """Get the correct data from a ResultsObject or a DataFrame"""
        
        if isinstance(source, pd.DataFrame):
            return source
        else:
            return getattr(source, var)
    
    def _from_trajectory(self, category, variable, source, obj):
        """Generic initialization/scaling function"""
        
        estimator = getattr(self, obj)
        method = getattr(estimator, f'{category}_from_trajectory')
        
        if variable is None:
            vars_to_init = get_vars(estimator.model)
        
            _print(f'The vars_to_init: {vars_to_init}')
            for var in vars_to_init:
                if hasattr(source, var):    
                    _print(f'Updating variable: {var}')
                    method(var, self._get_source_data(source, var))   
        else:
            method(variable, self._get_source_data(source, variable))
        return None
                                               
    def set_known_absorbing_species(self, *args, **kwargs):
        """Wrapper for set_known_absorbing_species in TemplateBuilder
        
        """
        self.builder.set_known_absorbing_species(*args, **kwargs)    
        return None
    
    def scale(self):
        """Scale the model"""
        
        parameter_dict = self.parameters.as_dict(bounds=False)    
        scaled_parameter_dict, scaled_models_dict = scale_models(self.model,
                                                                 parameter_dict,
                                                                 name=self.name,
                                                                 )         
        return scaled_parameter_dict, scaled_models_dict
    
    def clone(self, *args, **kwargs):
        """Makes a copy of the ReactionModel and removes the data. This is done
        to reuse the model, components, and parameters in an easier manner
        
        """
        new_kipet_model = copy.deepcopy(self)
        
        name = kwargs.get('name', self.name + '_copy')
        copy_model = kwargs.get('model', True)
        copy_builder = kwargs.get('builder', True)
        copy_components = kwargs.get('components', True)   
        copy_parameters = kwargs.get('parameters', True)
        copy_datasets = kwargs.get('datasets', True)
        copy_constants = kwargs.get('constants', True)
        copy_settings = kwargs.get('settings', True)
        copy_algebraic_variables = kwargs.get('alg_vars', True)
        copy_odes = kwargs.get('odes', True)
        copy_algs = kwargs.get('algs', True)
        
        # Reset the datasets
        
        new_kipet_model.name = name
        
        if not copy_model:
            new_kipet_model.model = None
        
        if not copy_builder:
            new_kipet_model.builder = TemplateBuilder()
            
        if not copy_components:
            new_kipet_model.components = ComponentBlock()
        
        if not copy_parameters:
            new_kipet_model.parameters = ParameterBlock()
            
        if not copy_datasets:
            del new_kipet_model.datasets
            new_kipet_model.datasets = DataBlock()
            
        if not copy_constants:
            new_kipet_model.constants = None
            
        if not copy_algebraic_variables:
            new_kipet_model.algebraic_variables = []
            
        if not copy_settings:
            new_kipet_model.settings = Settings()
            
        if not copy_odes:
            new_kipet_model.odes = None
            
        if not copy_algs:
            new_kipet_model.algs = None
        
        list_of_attr_to_delete = ['p_model', 'v_model', 'p_estimator',
                                  'v_estimator', 'simulator']
        
        for attr in list_of_attr_to_delete:
            if hasattr(new_kipet_model, attr):
                setattr(new_kipet_model, attr, None)
        
        new_kipet_model.results_dict = {}
            
        return new_kipet_model
    
    def rhps_method(self,
                     method='k_aug',
                     calc_method='global',
                     scaled=True):
        """This calls the reduce_models method in the EstimationPotential
        module to reduce the model based on the reduced hessian parameter
        selection method.
        
        Args:
            kwargs:
                replace (bool): defaults to True, option to replace the
                    parameters deemed unestimable from the model with constants
                no_scaling (bool): defaults to True, removes the scaling
                    constants from the model and restores the parameter values
                    and their bounds.
                    
        Returns:
            results (ResultsObject): A standard results object with the reduced
                model results
        
        """
        if self.model is None:
            self.create_pyomo_model()
            
        kwargs = {}
        kwargs['solver_opts'] = self.settings.solver
        kwargs['method'] = method
        kwargs['calc_method'] = calc_method
        kwargs['scaled'] = scaled
        kwargs['use_bounds'] = False
        kwargs['use_duals'] = False
        kwargs['ncp'] = self.settings.collocation.ncp
        kwargs['nfe'] = self.settings.collocation.nfe
        
        # parameter_dict = self.parameters.as_dict(bounds=True)
        results, reduced_model = rhps_method(self.model, **kwargs)
        
#        results.file_dir = self.settings.general.charts_directory
        
        self.reduced_model = reduced_model
        self.using_reduced_model = True
        self.reduced_model_results = results
        
        #Make a KipetModel as the result using the reduced model
        red_model = ReactionModel()
        self.red_model = ReactionModel(name=self.name+'_reduced')
                
        assign_list = ['components', 'parameters', 'constants', 'algebraics',
                       'states', 'ub', 'settings', 'c', 'odes_dict']
  
        ignore = []
        for item in assign_list:
            if item not in ignore and hasattr(self, item):
                setattr(self.red_model, item, getattr(self, item))
        
        # reduced_kipet_model = self.clone('reduced_model', **items_not_copied)
        
        #%%
        # reduced_kipet_model.add_parameter()
        # self, name=None, init=None, bounds=None
        
        # reduced_kipet_model.model = reduced_model
        # reduced_kipet_model.results = results
        
        # reduced_parameter_set = {k: [v.value, (v.lb, v.ub)] for k, v in reduced_kipet_model.model.P.items()}
        # for param, param_data in reduced_parameter_set.items():
        #     reduced_kipet_model.add_parameter(param, init=param_data[0], bounds=param_data[1])
        
        # return reduced_kipet_model
    
    def set_non_absorbing_species(self, non_abs_list):
        """Wrapper for set_non_absorbing_species in TemplateBuilder"""
        
        self._has_non_absorbing_species = True
        self.non_abs_list = non_abs_list
        return None
        
    def add_noise_to_data(self, var, noise, overwrite=False):
        """Wrapper for adding noise to data after data has been added to
        the specific ReactionModel
        
        """
        dataframe = self.datasets[var].data
        if overwrite:
            self.datasets[var].data = dataframe
        return data_tools.add_noise_to_signal(dataframe, noise)    
    
    def unwanted_contribution(self, variant, **kwargs):
        
        self._G_contribution = variant
        self._G_data = {'G_contribution': variant,
                        'Z_in': kwargs.get('Z_in', dict()),
                        'St': kwargs.get('St', dict()),
                        }
    
    def analyze_parameters(self, 
                        method=None,
                        parameter_uncertainties=None,
                        meas_uncertainty=None,
                        sigmas=None,
                        ):
        
        """This is a wrapper for the EstimabilityAnalyzer 
        """
        # Here we use the estimability analysis tools
        self.e_analyzer = EstimabilityAnalyzer(self.model)
        # Problem needs to be discretized first
        self.e_analyzer.apply_discretization('dae.collocation',
                                             nfe=60,
                                             ncp=1,
                                             scheme='LAGRANGE-RADAU')
        
        #param_uncertainties = {'k1':0.09,'k2':0.01,'k3':0.02,'k4':0.5}
        # sigmas, as before, represent the variances in regard to component
        #sigmas = {'A':1e-10,'B':1e-10,'C':1e-11, 'D':1e-11,'E':1e-11,'device':3e-9}
        # measurement scaling
        #meas_uncertainty = 0.05
        # The rank_params_yao function ranks parameters from most estimable to least estimable 
        # using the method of Yao (2003). Notice the required arguments. Returns a dictionary of rankings.
        if method == 'yao':
            
            listparams = self.e_analyzer.rank_params_yao(meas_scaling=meas_uncertainty,
                                                         param_scaling=parameter_uncertainties,
                                                         sigmas=sigmas)
            #print(listparams)
            
            # Now we can run the analyzer using the list of ranked parameters
            params_to_select = self.e_analyzer.run_analyzer(method='Wu', 
                                                            parameter_rankings=listparams,
                                                            meas_scaling=meas_uncertainty, 
                                                            variances=sigmas
                                                            )
            # We can then use this information to fix certain parameters and run the parameter estimation
            print(params_to_select)
            
            params_to_fix = list(set(self.parameters.names).difference(params_to_select))
        
        return params_to_select, params_to_fix 
    
    def fix_and_remove_parameters(self, model_name, parameters=None):
        
        if model_name not in ['s_model', 'v_model', 'p_model']:
            raise ValueError(f'ReactionModel does not have model type {model_name}')
        
        model = getattr(self, model_name)
        param_replacer = ParameterReplacer([model], fix_parameters=parameters)
        param_replacer.remove_fixed_vars()
    
        return None

        
    """MODEL FUNCTION AREA"""
    
    def make_reaction_table(self, stoich_coeff, rxns):
        
        df = pd.DataFrame()
        for k, v in stoich_coeff.items():
            df[k] = v
        
        df.index = rxns
        self.reaction_table = df
    
    def reaction_block(self, stoich_coeff, rxns):
        # make a reactions_dict in __init__ and update here
        """Method to allow for simple construction of a reaction system
        
        Args:
            stoich_coeff (dict): dict with components as keys and stoichiometric
                coeffs is lists as values:
                    example: stoich_coeff['A'] = [-1, 0 , 1] etc
                
            rxns (list): list of algebraics that represent reactions
            
        Returns:
            r_dict (dict): dict of completed reaction expressions
            
        """
        self.make_reaction_table(stoich_coeff, rxns)
        
        r_dict = {}
        
        for com in self.components.names: 
            r_dict[com] = sum(stoich_coeff[com][i] * self.ae(r) for i, r in enumerate(rxns)) 
    
        return r_dict
    
    @property
    def models(self):
        
        output = 'ReactionModel has the following:\n'
        output_dict = {}
        
        for model in [name + 'model' for name in ['', 's_', 'v_', 'p_']]:
        
            if hasattr(self, model):
                output += f'{model} True\n'
                output_dict[model] = True
            else:
                output += f'{model} False\n'
                output_dict[model] = False
            
        print(output)
        return output_dict
    
    @property
    def has_objective(self):
        """Check if p_model has an objective"""
        
        return hasattr(self.p_model, 'objective')


    def check_component_units(self):
        """Method to check whether the units provided are consisten with the
        base units
        
        """
        
        print('Checking model component units:\n')
        
        element_dict = {
           'parameters': self.parameters,
           'components': self.components,
           'constants': self.constants,
           'algebraics': self.algebraics,
           'states': self.states,
                }
        
        if not hasattr(self, 'c'):
            self._make_c_dict()
        
        for elem, obj in element_dict.items():
            for comp in obj:
                comp._check_scaling()
                
                if comp.units != comp.units_orig:
                    comp.pyomo_var.parent_component()._units = getattr(pyo_units, str(comp.units.u))
        
        self._units_checked = True
        print('')
        
        return None


    def check_component_units_base(self):
        """Method to check whether the units provided are consisten with the
        base units
        
        """
        element_dict = {
           'parameters': self.parameters,
           'components': self.components,
           'constants': self.constants,
           'algebraics': self.algebraics,
           'states': self.states,
                }
        
        from kipet.model_components.units_handler import convert_single_dimension
        
        if not hasattr(self, 'c'):
            self._make_c_dict()
        
        for elem, obj in element_dict.items():
            if elem in ['components', 'states', 'parameters', 'constants']:
                for key in obj:
                    
                    print(f'Checking units for {key.name}: {key.units}')
                    
                    key_comp = key #self.get_state(key)
                    key_comp_units = key_comp.units
                    key_comp_units = convert_single_dimension(self.unit_base.ur, key_comp_units, self.unit_base.TIME_BASE, power_fixed=False)
                    
                    print(key_comp_units)
                    print(self.unit_base.VOLUME_BASE)
                    print(f'Checking units for {key.name}: {key.units}')
                    key_comp_units = convert_single_dimension(self.unit_base.ur, key_comp_units, self.unit_base.VOLUME_BASE, power_fixed=True)
                    
                    print(key_comp_units)
                    
                    key_comp.units = key_comp_units.units
                    key_comp.value *= key_comp_units.m
                    key_comp.pyomo_var.parent_component()._units = getattr(pyo_units, str(key_comp.units))
        
                    # if self.value is not None:
                    #     self.value = quantity.m*self.value
                        
                    key_comp.conversion_factor = key_comp_units.m
                    # key_comp.units = 1*quantity.units
                    
                    if hasattr(key_comp, 'bounds') and key_comp.bounds is not None:
                        bounds = list(key_comp.bounds)
                        if bounds[0] is not None:
                            bounds[0] *= key_comp.conversion_factor
                        if bounds[1] is not None:
                            bounds[1] *= key_comp.conversion_factor
                        key_comp.bounds = (bounds) 
        
        # for elem, obj in element_dict.items():
        #     for comp in obj:
        #         comp._check_scaling()
                
        #         if comp.units != comp.units_orig:
        #             comp.pyomo_var.parent_component()._units = getattr(pyo_units, str(comp.units.u))
        
        self._units_checked = True
        print('')
        
        return None
    
    def check_model_units(self, orig_units=False, display=False):
        """Method to check the expected units of the algebraic and odes
        
        """
        from kipet.model_components.units_handler import convert_single_dimension
        
        print('Checking expected equation units:\n')
        
        if not self.__flag_odes_built:
            self._build_odes()
        if not self.__flag_algs_built:
            self._build_algs()
        
        elements = ['parameters', 'components', 'states', 'constants', 'algebraics']
        element_dict = {}
        for element in elements:
            if hasattr(self, f'{element}'):
                element_dict[element] = getattr(self, f'{element}')
            else:
                element_dict[element] = {}
                
        if not hasattr(self, 'c'):
            self._make_c_dict()
        
        for key, expr in self.odes_dict.items():
            
            key_comp = self.get_state(key)
            if orig_units:
                convert_to = ' / '.join([str(key_comp.units_orig), self.unit_base.TIME_BASE])    
                key_comp.use_orig_units = True
            else:
                convert_to = ' / '.join([str(key_comp.units), self.unit_base.TIME_BASE])
            
            expr.check_expression_units(convert_to=str(convert_to), scalar=1)
        
            
        for key, expr in self.algs_dict.items():
            
            expr_obj = self.alg_obj.exprs[key]
            
            if expr_obj.expression_orig is not None:
                expr_use = expr_obj.expression_orig
            else:
                expr_use = expr_obj.expression
            
            print(expr_use)
            key_comp = pyo_units.get_units(expr_use)
            expr.units = key_comp
            
        if display:
            self.ode_obj.display_units()
            print('')
            if len(self.alg_obj) > 0:
                self.alg_obj.display_units()
                print('')
            
            
        return None
    
    
    def plot(self, var=None, jupyter=False):
        
        """Plot results using the variable or variable class"""
        
        from kipet.visuals.plots import PlotObject
        
        self._plot_object = PlotObject(reaction_model=self, jupyter=jupyter)
        
        if var == 'Z':
            self._plot_object._plot_all_Z()
            
        if var == 'S':
            self._plot_object._plot_all_S()
            
        elif var in self.components.names:
            self._plot_object._plot_Z(var)           
            
        elif var in self.states.names:
            self._plot_object._plot_X(var)

        elif var in self.algebraics.names:
            self._plot_object._plot_Y(var)            
                    
        elif var is None:
            self._plot_object._plot_all_Z()
            
            if hasattr(self.results, 'S'):
                self._plot_object._plot_all_S()
            
            for var in self.states.names:
                self._plot_object._plot_X(var)
            
            for var in self.algebraics.names:
                self._plot_object._plot_Y(var)  
                
            if hasattr(self.results, 'step'):
                for var in self.results.step:
                    self._plot_object._plot_step(var)
                
        return None
        

    def ae(self, alg_var):
        return self.algs_dict[alg_var].expression
    
    def ode(self, ode_var):
        return self.odes_dict[ode_var].expression

    def get_state(self, comp):
        if comp in self.components:
            return self.components[comp]
        elif comp in self.states:
            return self.states[comp]
 
    def get_alg(self, comp):
        if comp in self.algebraics:
            return self.algebraics[comp]
        
    
    
def _set_directory(model_object, filename, abs_dir=False):
    """Wrapper for the set_directory method. This replaces the awkward way
    of ensuring the correct directory for the data is used.
    
    Args:
        filename (str): the file name to be formatted
        
    Returns:
        file_path (pathlib Path): The absolute path of the given file
    """
    #directory = model_object.settings.general.data_directory
    directory = pathlib.Path.cwd()
    file_path = pathlib.Path(directory).joinpath(filename)
    
    return file_path


from functools import reduce
import sys
import warnings

import pandas as pd
from pyomo.dae import *
from pyomo.environ import *

from kipet.core_methods.ResultsObject import *
from kipet.core_methods.Simulator import *
from kipet.post_model_build.pyomo_model_tools import (
    index_set_info,
    get_index_sets,
    )
from kipet.common.VisitorClasses import ScalingVisitor
from kipet.dev_tools.display import Print

from kipet.top_level.variable_names import VariableNames

__var = VariableNames()
DEBUG=__var.DEBUG
_print = Print(verbose=DEBUG)

class PyomoSimulator(Simulator):
    """Simulator based on pyomo.dae discretization strategies.

    Attributes:
        model (Pyomo model)

        _times (array_like): array of times after discretization

        _n_times (int): number of discretized time points

        _ipopt_scaled (bool): flag that indicates if there are
        ipopt scaling factors specified
    """

    def __init__(self, model):
        """Simulator constructor.

        Note:
            Makes a shallow copy to the model. Changes applied to
            the model within the simulator are applied to the original
            model passed to the simulator

        Args:
            model (Pyomo model)
        """
        self.__var = VariableNames()

        super(PyomoSimulator, self).__init__(model)
        self._alltimes = sorted(self.model.alltime)#added for special structure CS
        #self._times = sorted(self.model.time)
        self._n_alltimes = len(self._alltimes) #added for special structure CS
        #self._n_times = len(self._times)
        self._meas_times=sorted(self.model.meas_times)
        self._allmeas_times=sorted(self.model.allmeas_times)
        self._ipopt_scaled = False
        self._spectra_given = hasattr(self.model, self.__var.spectra_data)
        self._concentration_given = hasattr(self.model, self.__var.concentration_measured) or hasattr(self.model, self.__var.user_defined) or hasattr(self.model, self.__var.state)
        self._conplementary_states_given = hasattr(self.model, self.__var.state)
        self._absorption_given = hasattr(self.model, self.__var.spectra_species)  # added for special case of absorption data available but not concentration data CS
        self._huplc_given = hasattr(self.model, 'Chat')
        self._smoothparam_given = hasattr(self.model, self.__var.smooth_parameter)

        # creates scaling factor suffix
        if not hasattr(self.model, 'scaling_factor'):
            self.model.scaling_factor = Suffix(direction=Suffix.EXPORT)

    def apply_discretization(self, transformation, **kwargs):
        """Discretizes the model.

        Args:
            transformation (str): TODO
            same keywords as in pyomo.dae method

        Returns:
            None
        """
        # Additional keyword - not to use alltimes as the basis for the finite elements.
        
        fixed_times = kwargs.pop('fixed_times', None)
        
        if not self.model.alltime.get_discretization_info():
            
            discretizer = TransformationFactory(transformation)

            if fixed_times == None:
                discretizer.apply_to(self.model, wrt=self.model.alltime, **kwargs)
            else:
                discretizer.apply_to(self.model, wrt=fixed_times, **kwargs)
                
            self._alltimes = sorted(self.model.alltime)
            self._n_alltimes = len(self._alltimes)

            #added for optional smoothing parameter with reading values from file CS:
            # if self._smoothparam_given:
            #     dfps = pd.DataFrame(index=self.model.alltime, columns=self.model.smoothparameter_names)
            #     for t in self.model.alltime:
            #         if t not in self.model.allsmooth_times:  # for points that are the same in original meas times and feed times
            #             dfps.loc[t] = float(22.) #something that is not between 0 and 1
            #         else:
            #             ps_dict_help = dict()
            #             for p in self.model.smoothparameter_names:
            #                 ps_dict_help[t, p] = value(self.model.smooth_param_data[t, p])
            #             dfps.loc[t] = [ps_dict_help[t, p] for p in self.model.smoothparameter_names]
            #     dfallps = dfps
            #     dfallps.sort_index(inplace=True)
            #     dfallps.index = dfallps.index.to_series().apply(
            #         lambda x: np.round(x, 6))  # time from data rounded to 6 digits

            #     dfallpsall = pd.DataFrame(index=self.model.alltime, columns=self.model.smoothparameter_names)
            #     dfsmoothdata = pd.DataFrame(index=sorted(self.model.smooth_param_datatimes), columns=self.model.smoothparameter_names)

            #     for t in self.model.smooth_param_datatimes:
            #         dfsmoothdata.loc[t] = [value(self.model.smooth_param_data[t, p]) for p in self.model.smoothparameter_names]

            #     for p in self.model.smoothparameter_names:
            #         values = interpolate_trajectory(self.model.alltime, dfsmoothdata[p])
            #         for i, ti in enumerate(self.model.alltime):
            #             if float(dfallps[p][ti]) > 1:
            #                 valueinterp=values[i]
            #                 dfallpsall[p][ti] = float(valueinterp)
            #             else:
            #                 dfallpsall.loc[ti] = float(dfallps[p][ti])

            self._default_initialization()
            
            # if hasattr(self.model, 'K'):
            #     print('Scaling the parameters')
            #     self.scale_parameters()
            
        else:
            print('***WARNING: Model already discretized. Ignoring second discretization')
            
    def scale_model(self):
        if hasattr(self.model, self.__var.model_parameter_scaled):
            print('Scaling the parameters')
            self.scale_parameters()
            
        return None
    
    def scale_parameters(self):
        """If scaling, this multiplies the constants in model.K to each
        parameter in model.P.
        
        I am not sure if this is necessary and will look into its importance.
        """
        #if self.model.K is not None:
        self.scale = {}
        for i in self.model.P:
            self.scale[id(self.model.P[i])] = self.model.K[i]

        for var in self.__var.modeled_states:

            for i in getattr(self.model, var):
                self.scale[id(getattr(self.model, var)[i])] == 1

        # for i in self.model.Z:
        #     self.scale[id(self.model.Z[i])] = 1
            
        # for i in self.model.dZdt:
        #     self.scale[id(self.model.dZdt[i])] = 1
            
        # for i in self.model.X:
        #     self.scale[id(self.model.X[i])] = 1
    
        # for i in self.model.dXdt:
        #     self.scale[id(self.model.dXdt[i])] = 1
    
        for k, v in getattr(self.model, self.__var.ode_constraints).items():
        # for k, v in self.model.odes.items(): 
            scaled_expr = self.scale_expression(v.body, self.scale)
            # self.model.odes[k] = scaled_expr == 0
            getattr(self.model, self.__var.ode_constraints)[k] = scaled_expr == 0
    
    def scale_expression(self, expr, scale):
        
        visitor = ScalingVisitor(scale)
        return visitor.dfs_postorder_stack(expr)

    def fix_from_trajectory(self, variable_name, variable_index, trajectories, verbose=False):

        if variable_name in  self.__var.modeled_states:
            
            raise NotImplementedError("Fixing state variables is not allowd. Only algebraics can be fixed")

        single_traj = trajectories[variable_index]
        sim_alltimes = sorted(self._alltimes)
        var = getattr(self.model, variable_name)
        values = interpolate_trajectory(sim_alltimes, single_traj)
        r_values = []
        for i, t in enumerate(sim_alltimes):
            var[t, variable_index].fix(values[i])
            r_values.append((t, values[i]))

        return r_values

    def unfix_time_dependent_variable(self, variable_name, variable_index):
        var = getattr(self.model, variable_name)
        sim_times = sorted(self._alltimes)
        for i, t in enumerate(sim_times):
            var[t, variable_index].fixed = False

    # initializes the trajectories to the initial conditions
    def _default_initialization(self):
        """Initializes discreted variables model with initial condition values.

           This method is not intended to be used by users directly
        Args:
            None

        Returns:
            None
        """
        tol = 1e-4
        
        if hasattr(self.model, self.__var.concentration_model):
            z_init = []
            for t in self._alltimes:
                for k in self._mixture_components:
                    if abs(self.model.init_conditions[k].value) > tol:
                        z_init.append(self.model.init_conditions[k].value)
                    else:
                        z_init.append(1.0)
    
            z_array = np.array(z_init).reshape((self._n_alltimes, self._n_components))
            z_init_panel = pd.DataFrame(data=z_array,
                                        columns=self._mixture_components,
                                        index=self._alltimes)
            
            self.initialize_from_trajectory(self.__var.concentration_model, z_init_panel)

        c_init = []
        if self._concentration_given:
            pass
        else:
            for t in self._allmeas_times:
                for k in self._mixture_components:
                    if t in self._meas_times:
                        if abs(self.model.init_conditions[k].value) > tol:
                            c_init.append(self.model.init_conditions[k].value)
                        else:
                            c_init.append(1.0)
                    else: c_init.append(float('nan')) #added for new huplc structure!

        if self._n_allmeas_times:
            if self._concentration_given:
                pass
            else:
                c_array = np.array(c_init).reshape((self._n_allmeas_times, self._n_components))
                c_init_panel = pd.DataFrame(data=c_array,
                                        columns=self._mixture_components,
                                        index=self._allmeas_times)

                if hasattr(self.model, self.__var.concentration_measured):
                    self.initialize_from_trajectory(self.__var.concentration_measured, c_init_panel)
                    print("self._n_meas_times is true in _default_init in PyomoSim")

        if hasattr(self.model, self.__var.state_model):
            x_init = []
            for t in self._alltimes:
                for k in self._complementary_states:
                    if abs(self.model.init_conditions[k].value) > tol:
                        x_init.append(self.model.init_conditions[k].value)
                    else:
                        x_init.append(1.0)
    
            x_array = np.array(x_init).reshape((self._n_alltimes, self._n_complementary_states))
            x_init_panel = pd.DataFrame(data=x_array,
                                        columns=self._complementary_states,
                                        index=self._alltimes)

            self.initialize_from_trajectory(self.__var.state_model, x_init_panel)

    def initialize_parameters(self, params):
        for k, v in params.items():
            getattr(self.model, self.__var.model_parameter).value = v

    
    def build_sets_new(self, variable_name, trajectories):
        
        var = getattr(self.model, variable_name)
        index_sets = get_index_sets(var)
        
        if isinstance(trajectories, pd.DataFrame):
            if variable_name not in self.__var.__dict__.values():
                _print(f'Update of {variable_name} (non KIPET standard)')
                inner_set = list(trajectories.index)
                component_set = list(trajectories.columns)
            else:
                _print(f'Update of KIPET Var {variable_name}')
                
                if var.dim() > 1 and var.index_set().dim() == 0:
                    inner_set = list(set([t[0] for t in var.index_set()]))
                    component_set = list(set([t[1] for t in var.index_set()]))
                    inner_set.sort()
                else:
                    index_sets = get_index_sets(var)
                    index_dict = index_set_info(index_sets)
                    
                    inner_set = getattr(self.model, index_sets[index_dict['cont_set'][0]].name)
                    component_set = getattr(self.model, index_sets[index_dict['other_set'][0]].name)
                
            return inner_set, component_set
                
        else:
            _print('Unsupported data type for initialization...')
            return None, None
        
    
    # def build_sets(self, variable_name, trajectories):
        
    #     if isinstance(trajectories, pd.DataFrame):
    #         if variable_name not in set_comp: # and isinstance(trajectories, dict):
    #             _print(f'Update of {variable_name} (non KIPET standard)')
    #             inner_set = list(trajectories.index)
    #             component_set = list(trajectories.columns)
    #         else:
    #             _print(f'Update of KIPET Var {variable_name}')
    #             inner_set = rgetattr(self, set_time[variable_name])
    #             component_set = rgetattr(self, set_comp[variable_name])
            
    #         if variable_name == 'Cm':
    #             print(inner_set)
            
    #         return inner_set, component_set
            
    #     else:
    #         _print('Unsupported data type for initialization...')
    #         return None, None
        
    def initialize_from_trajectory(self, variable_name, trajectories):
        """Initializes discretized points with values from trajectories.
        Args:
            variable_name (str): Name of the variable in pyomo model
            trajectories (DataFrame or Series): Indexed in in the same way the pyomo
            variable is indexed. If the variable is by two sets then the first set is
            the indices of the data frame, the second set is the columns
        Returns:
            None
        """
        _print(f'Initialization of Var: {variable_name}')
        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before initializing')
            
        var = getattr(self.model, variable_name)
        inner_set, component_set = self.build_sets_new(variable_name, trajectories)
       
        if inner_set is None and component_set is None:
            return None

        for component in component_set:
            if component in trajectories.columns:
                single_trajectory = trajectories[component]
                values = interpolate_trajectory(inner_set, single_trajectory)
                for i, t in enumerate(inner_set):
                    if not np.isnan(values[i]):
                        var[t, component].value = values[i]

        return None

    def scale_variables_from_trajectory(self, variable_name, trajectories):
        """Scales discretized variables with maximum value of the trajectory.
        Note:
            This method only works with ipopt
        Args:
            variable_name (str): Name of the variable in pyomo model
            trajectories (DataFrame or Series): Indexed in in the same way the pyomo
            variable is indexed. If the variable is by two sets then the first set is
            the indices of the data frame, the second set is the columns
        Returns:
            None
        """
        tol = 1e-5
        
        var = getattr(self.model, variable_name)
        inner_set, component_set = self.build_sets_new(variable_name, trajectories)
       
        if inner_set is None and component_set is None:
            return None
        
        for component in component_set:
            
            if component not in component_set:
                raise RuntimeError(f'Component {component} is not used in the model')
                
            nominal_vals = abs(trajectories[component].max())
            if nominal_vals >= tol:
                scale = 1.0 / nominal_vals
                for t in inner_set:
                    self.model.scaling_factor.set_value(var[t, component], scale)

        self._ipopt_scaled = True
        return None
        

    def run_sim(self, solver, **kwds):
        """ Runs simulation by solving nonlinear system with ipopt

        Args:
            solver (str): name of the nonlinear solver to used

            solver_opts (dict, optional): Options passed to the nonlinear solver

            variances (dict, optional): Map of component name to noise variance. The
            map also contains the device noise variance

            tee (bool,optional): flag to tell the simulator whether to stream output
            to the terminal or not

        Returns:
            None

        """
        solver_opts = kwds.pop('solver_opts', dict())
        sigmas = kwds.pop('variances', dict())
        tee = kwds.pop('tee', False)
        seed = kwds.pop('seed', None)

        if not self.model.alltime.get_discretization_info():
            raise RuntimeError('apply discretization first before runing simulation')

        # adjusts the seed to reproduce results with noise
        np.random.seed(seed)
        
        # variables
        Z_var = self.model.Z
        #dZ_var = self.model.dZdt
        #P_var = self.model.P
        #X_var = self.model.X
        #U_var = self.model.U
        #dX_var = self.model.dXdt
        if hasattr(self.model, 'Cm'):
            C_var = self.model.Cm  # added for estimation with inputs and conc data CS
        if self._huplc_given: #added for additional data CS
            Dhat_var = self.model.Dhat
            Chat_var = self.model.Chat
        # check all parameters are fixed before simulating
        # for p_var_data in P_var.values():
        #     if not p_var_data.fixed:
        #         raise RuntimeError(
        #             'For simulation fix all parameters. Parameter {} is unfixed'.format(p_var_data.getname()))

        # deactivates objective functions for simulation
        if self.model.nobjectives():
            objectives_map = self.model.component_map(ctype=Objective, active=True)
            active_objectives_names = []
            for obj in objectives_map.values():
                name = obj.getname()
                active_objectives_names.append(name)
                str_warning = 'Deactivating objective {} for simulation'.format(name)
                warnings.warn(str_warning)
                obj.deactivate()

        # Look at the output in results
        # self.model.write('f.nl')
        opt = SolverFactory(solver)

        for key, val in solver_opts.items():
            opt.options[key] = val
            
        solver_results = opt.solve(self.model, tee=tee, symbolic_solver_labels=True)
        results = ResultsObject()

        # activates objective functions that were deactivated
        if self.model.nobjectives():
            active_objectives_names = []
            objectives_map = self.model.component_map(ctype=Objective)
            for name in active_objectives_names:
                objectives_map[name].activate()

        # retriving solutions to results object
        results.load_from_pyomo_model(self.model,
                                      to_load=None)

        # c_noise_results = []

        # w = np.zeros((self._n_components, self._n_allmeas_times))
        # n_sig = np.zeros((self._n_components, self._n_allmeas_times))
        # # for the noise term
        # if sigmas:
        #     for i, k in enumerate(self._mixture_components):
        #         if k in sigmas.keys():
        #             sigma = sigmas[k] ** 0.5
        #             dw_k = np.random.normal(0.0, sigma, self._n_allmeas_times)
        #             n_sig[i, :] = np.random.normal(0.0, sigma, self._n_allmeas_times)
        #             w[i, :] = np.cumsum(dw_k)

        # # this addition is not efficient but it can be changed later

        # for i, t in enumerate(self._allmeas_times):
        #     for j, k in enumerate(self._mixture_components):
        #         # c_noise_results.append(Z_var[t,k].value+ w[j,i])
        #         c_noise_results.append(Z_var[t, k].value + n_sig[j, i])

        # c_noise_array = np.array(c_noise_results).reshape((self._n_allmeas_times, self._n_components))
        # results.C = pd.DataFrame(data=c_noise_array,
        #                           columns=self._mixture_components,
        #                           index=self._allmeas_times)

        # #added due to new structure for non_abs species, Cs as subset of C (CS):
        # if hasattr(self, '_abs_components'):
        #     cs_noise_results=[]
        #     for i, t in enumerate(self._allmeas_times):
        #         # if i in self._meas_times:
        #         for j, k in enumerate(self._abs_components):
        #             # c_noise_results.append(Z_var[t,k].value+ w[j,i])
        #             cs_noise_results.append(Z_var[t, k].value + n_sig[j, i])

        #     cs_noise_array = np.array(cs_noise_results).reshape((self._n_allmeas_times, self._nabs_components))
        #     results.Cs = pd.DataFrame(data=cs_noise_array,
        #                               columns=self._abs_components,
        #                               index=self._allmeas_times)

#        addition for inputs estimation with concentration data CS:
        # if self._concentration_given == True and self._absorption_given == False:
        #     c_noise_results = []
        #     for i, t in enumerate(self._allmeas_times):
        #         # if i in self._meas_times:
        #         for j, k in enumerate(self._mixture_components):
        #             c_noise_results.append(C_var[t, k].value)
        #     c_noise_array = np.array(c_noise_results).reshape((self._n_allmeas_times, self._n_components))
        #     results.C = pd.DataFrame(data=c_noise_array,
        #                               columns=self._mixture_components,
        #                               index=self._allmeas_times)

        # if self._huplc_given == True:
        #     results.load_from_pyomo_model(self.model,
        #                               to_load=['Chat'])
        # s_results = []
        # # added due to new structure for non_abs species, non-absorbing species not included in S (CS):
        # if hasattr(self, '_abs_components'):
        #     for l in self._meas_lambdas:
        #         for k in self._abs_components:
        #             s_results.append(self.model.S[l, k].value)
        # else:
        #     for l in self._meas_lambdas:
        #         for k in self._mixture_components:
        #             s_results.append(self.model.S[l, k].value)

        # d_results = []
        # if sigmas:
        #     sigma_d = sigmas.get('device') ** 0.5 if "device" in sigmas.keys() else 0
        # else:
        #     sigma_d = 0
        # if s_results and c_noise_results:
        #     # added due to new structure for non_abs species, Cs and S as above(CS):
        #     if hasattr(self,'_abs_components'):
        #         for i, t in enumerate(self._meas_times):
        #             #if t in self._meas_times:
        #                 # print(i, t)
        #             for j, l in enumerate(self._meas_lambdas):
        #                 suma = 0.0
        #                 for w, k in enumerate(self._abs_components):
        #                     # print(i, self._meas_times)
        #                     Cs = cs_noise_results[i * self._nabs_components + w]
        #                     S = s_results[j * self._nabs_components + w]
        #                     suma += Cs * S
        #                 if sigma_d:
        #                     suma += np.random.normal(0.0, sigma_d)
        #                 d_results.append(suma)
        #                 # print(d_results)
        #     else:
        #         for i, t in enumerate(self._meas_times):
        #             # # print(i, t)
        #             # if t in self._meas_times:
        #             for j, l in enumerate(self._meas_lambdas):
        #                 suma = 0.0
        #                 for w, k in enumerate(self._mixture_components):
        #                     # print(i, self._meas_times)
        #                     C = c_noise_results[i * self._n_components + w]
        #                     S = s_results[j * self._n_components + w]
        #                     suma += C * S
        #                 if sigma_d:
        #                     suma += np.random.normal(0.0, sigma_d)
        #                 d_results.append(suma)
        #                 # print(d_results)
        # # added due to new structure for non_abs species, non-absorbing species not included in S (CS):
        # if hasattr(self, '_abs_components'):
        #     s_array = np.array(s_results).reshape((self._n_meas_lambdas, self._nabs_components))
        #     results.S = pd.DataFrame(data=s_array,
        #                               columns=self._abs_components,
        #                               index=self._meas_lambdas)
        # else:
        #     s_array = np.array(s_results).reshape((self._n_meas_lambdas, self._n_components))
        #     results.S = pd.DataFrame(data=s_array,
        #                               columns=self._mixture_components,
        #                               index=self._meas_lambdas)


        # d_array = np.array(d_results).reshape((self._n_meas_times, self._n_meas_lambdas))
        # results.D = pd.DataFrame(data=d_array,
        #                          columns=self._meas_lambdas,
        #                          index=self._meas_times)

        # s_data_dict = dict()
        # for t in self._meas_times:
        #     for l in self._meas_lambdas:
        #         s_data_dict[t, l] = float(results.D[l][t])

        # # Added due to estimation with fe-factory and inputs where data already loaded to model before (CS)
        # if self._spectra_given:
        #     self.model.del_component(self.model.D)
        #     self.model.del_component(self.model.D_index)
        # #########

        # self.model.D = Param(self._meas_times,
        #                      self._meas_lambdas,
        #                      initialize=s_data_dict)

        # param_vals = dict()
        # for name in self.model.parameter_names:
        #     param_vals[name] = self.model.P[name].value

        # results.P = param_vals
        return results
    
# def rgetattr(obj, attr, *args):
#     def _getattr(obj, attr):
#         return getattr(obj, attr, *args)
#     return reduce(_getattr, [obj] + attr.split('.')) 